"""Tests for scrapy.core.downloader.handlers._httpx.HttpxDownloadHandler."""

from __future__ import annotations

import sys
from importlib.util import find_spec
from typing import TYPE_CHECKING, Any, ClassVar

import pytest

from scrapy import Request
from scrapy.core.downloader.handlers._httpx import (  # type: ignore[attr-defined]
    HAS_HTTP2,
    HAS_SOCKS,
    HttpxDownloadHandler,
    httpx,
)
from scrapy.exceptions import DownloadFailedError
from scrapy.utils.misc import build_from_crawler
from scrapy.utils.test import get_crawler
from tests import IDNA_REJECTED_HOSTNAMES
from tests.utils.bases.download_handlers_http import (
    TestHttpBase,
    TestHttpProxyBase,
    TestHttpsBase,
    TestHttpsCustomCiphersBase,
    TestHttpsDefaultCiphersBase,
    TestHttpsInvalidDNSIdBase,
    TestHttpsInvalidDNSPatternBase,
    TestHttpsTLSVersionBase,
    TestHttpsWrongHostnameBase,
    TestHttpWithCrawlerBase,
    TestMitmProxyBase,
    TestRealWebsiteBase,
    TestSimpleHttpsBase,
)
from tests.utils.decorators import coroutine_test

if TYPE_CHECKING:
    from pathlib import Path

    from scrapy.core.downloader.handlers import DownloadHandlerProtocol
    from tests.mockserver.http import MockServer
    from tests.mockserver.proxy_echo import ProxyEchoMockServer


pytestmark = pytest.mark.only_asyncio

if find_spec("httpx2") is None and find_spec("httpx") is None:
    pytest.skip("Neither httpx2 nor httpx are installed", allow_module_level=True)

# httpx2 < 2.4.0 reads URL.host through idna.decode() without display=True when
# building the Host header, which raises for punycode labels that IDNA 2008
# rejects. https://github.com/pydantic/httpx2/pull/1018
# This check can be dropped once the httpx2 requirement is bumped to >= 2.4.0.
try:
    for _hostname in IDNA_REJECTED_HOSTNAMES:
        _ = httpx.URL(f"http://{_hostname}").host
except Exception:
    HTTPX_SUPPORTS_IDNA_REJECTED_HOSTNAMES = False
else:
    HTTPX_SUPPORTS_IDNA_REJECTED_HOSTNAMES = True


class HttpxDownloadHandlerMixin:
    @property
    def download_handler_cls(self) -> type[DownloadHandlerProtocol]:
        return HttpxDownloadHandler

    @property
    def settings_dict(self) -> dict[str, Any] | None:
        return {
            "DOWNLOAD_HANDLERS": {
                "http": "scrapy.core.downloader.handlers._httpx.HttpxDownloadHandler",
                "https": "scrapy.core.downloader.handlers._httpx.HttpxDownloadHandler",
            }
        }


class TestHttp(HttpxDownloadHandlerMixin, TestHttpBase):
    handler_supports_bindaddress_meta = False
    handler_bad_header_handling = "fail"
    handler_supports_idna_rejected_hostnames = HTTPX_SUPPORTS_IDNA_REJECTED_HOSTNAMES

    @pytest.mark.skipif(
        sys.platform == "darwin",
        reason="127.0.0.2 is not available on macOS by default",
    )
    @coroutine_test
    async def test_bind_address_port_warning(
        self, caplog: pytest.LogCaptureFixture, mockserver: MockServer
    ) -> None:
        request = Request(mockserver.url("/client-ip"))
        async with self.get_dh(
            {"DOWNLOAD_BIND_ADDRESS": ("127.0.0.2", 12345)}
        ) as download_handler:
            response = await download_handler.download_request(request)
        assert response.body == b"127.0.0.2"
        assert "DOWNLOAD_BIND_ADDRESS specifies a port (12345)" in caplog.text
        assert "Ignoring the port" in caplog.text


class TestHttps(HttpxDownloadHandlerMixin, TestHttpsBase):
    handler_supports_bindaddress_meta = False
    handler_bad_header_handling = "fail"
    handler_supports_idna_rejected_hostnames = HTTPX_SUPPORTS_IDNA_REJECTED_HOSTNAMES
    tls_log_message = "SSL connection to 127.0.0.1 using protocol TLSv1.3, cipher"

    @pytest.mark.skip(reason="The check is Twisted-specific")
    def test_verify_certs_deprecated(self) -> None:  # type: ignore[override]
        pass


@pytest.mark.skipif(not HAS_HTTP2, reason="No HTTP/2 support in HttpxDownloadHandler")
class TestHttp2(TestHttps):
    http2 = True
    handler_supports_http2_dataloss = False

    default_handler_settings: ClassVar[dict[str, Any]] = {
        "HTTPX_HTTP2_ENABLED": True,
    }

    @coroutine_test
    async def test_protocol(self, mockserver: MockServer) -> None:
        request = Request(mockserver.url("/host", is_secure=self.is_secure))
        async with self.get_dh() as download_handler:
            response = await download_handler.download_request(request)
        assert response.protocol == "HTTP/2"

    @coroutine_test
    async def test_data_loss_handling(self, mockserver: MockServer) -> None:
        request = Request(mockserver.url("/broken", is_secure=self.is_secure))
        async with self.get_dh() as download_handler:
            with pytest.raises(DownloadFailedError):
                await download_handler.download_request(request)


class TestSimpleHttps(HttpxDownloadHandlerMixin, TestSimpleHttpsBase):
    pass


class TestHttpsWrongHostname(HttpxDownloadHandlerMixin, TestHttpsWrongHostnameBase):
    pass


class TestHttpsInvalidDNSId(HttpxDownloadHandlerMixin, TestHttpsInvalidDNSIdBase):
    pass


class TestHttpsInvalidDNSPattern(
    HttpxDownloadHandlerMixin, TestHttpsInvalidDNSPatternBase
):
    pass


class TestHttpsCustomCiphers(HttpxDownloadHandlerMixin, TestHttpsCustomCiphersBase):
    pass


class TestHttpsDefaultCiphers(HttpxDownloadHandlerMixin, TestHttpsDefaultCiphersBase):
    pass


class TestHttpsTLSVersion(HttpxDownloadHandlerMixin, TestHttpsTLSVersionBase):
    pass


class TestHttpWithCrawler(HttpxDownloadHandlerMixin, TestHttpWithCrawlerBase):
    pass


class TestHttpsWithCrawler(TestHttpWithCrawler):
    is_secure = True


class TestHttpProxy(HttpxDownloadHandlerMixin, TestHttpProxyBase):
    expected_http_proxy_request_body = b"http://example.com/"


class TestHttpsProxy(TestHttpProxy):
    is_secure = True

    @coroutine_test
    async def test_keylog(
        self,
        monkeypatch: pytest.MonkeyPatch,
        proxy_mockserver: ProxyEchoMockServer,
        tmp_path: Path,
    ) -> None:
        keylog_file = tmp_path / "keylog"
        monkeypatch.setenv("SSLKEYLOGFILE", str(keylog_file))
        http_proxy = proxy_mockserver.url("", is_secure=True)
        request = Request("http://example.com", meta={"proxy": http_proxy})
        async with self.get_dh() as download_handler:
            response = await download_handler.download_request(request)
        assert response.body == self.expected_http_proxy_request_body
        assert keylog_file.read_text()


@pytest.mark.requires_mitmproxy
class TestMitmProxy(HttpxDownloadHandlerMixin, TestMitmProxyBase):
    handler_supports_socks = HAS_SOCKS


@pytest.mark.requires_internet
class TestRealWebsite(HttpxDownloadHandlerMixin, TestRealWebsiteBase):
    pass


@pytest.mark.parametrize(("concurrency", "expected"), [(16, 16), (0, None)])
@coroutine_test
async def test_pool_limits(concurrency: int, expected: int | None) -> None:
    crawler = get_crawler(settings_dict={"CONCURRENT_REQUESTS": concurrency})
    handler = build_from_crawler(HttpxDownloadHandler, crawler)
    try:
        assert handler._limits.max_connections == expected
        assert handler._limits.max_keepalive_connections == expected
    finally:
        await handler.close()



"""
This module contains data types used by Scrapy which are not included in the
Python Standard Library.
"""

from __future__ import annotations

import collections
import contextlib
import weakref
from collections import OrderedDict
from typing import TYPE_CHECKING, Any, TypeVar, cast

if TYPE_CHECKING:
    from collections.abc import Container

    # typing.Self requires Python 3.11
    from typing_extensions import Self


_KT = TypeVar("_KT")
_VT = TypeVar("_VT")


class CaseInsensitiveDict(collections.UserDict[str | bytes, Any]):
    """A dict-like structure that accepts strings or bytes
    as keys and allows case-insensitive lookups.
    """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        self._keys: dict[str | bytes, Any] = {}
        super().__init__(*args, **kwargs)

    def __getitem__(self, key: str | bytes) -> Any:
        normalized_key = self._normkey(key)
        return super().__getitem__(self._keys[normalized_key.lower()])

    def __setitem__(self, key: str | bytes, value: Any) -> None:
        normalized_key = self._normkey(key)
        try:
            lower_key = self._keys[normalized_key.lower()]
            del self[lower_key]
        except KeyError:
            pass
        super().__setitem__(normalized_key, self._normvalue(value))
        self._keys[normalized_key.lower()] = normalized_key

    def __delitem__(self, key: str | bytes) -> None:
        normalized_key = self._normkey(key)
        stored_key = self._keys.pop(normalized_key.lower())
        super().__delitem__(stored_key)

    def __contains__(self, key: str | bytes) -> bool:  # type: ignore[override]
        normalized_key = self._normkey(key)
        return normalized_key.lower() in self._keys

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}: {super().__repr__()}>"

    # UserDict.copy() shallow-copies the instance, which would share self._keys
    # between the copy and the original.
    def __copy__(self) -> Self:
        new = self.__class__()
        new.data = self.data.copy()
        new._keys = self._keys.copy()
        return new

    copy = __copy__

    # UserDict.__ior__ updates self.data directly, which would leave self._keys
    # out of date.
    def __ior__(self, other: Any) -> Self:  # type: ignore[override,misc]
        self.update(other)
        return self

    def _normkey(self, key: str | bytes) -> str | bytes:
        return key

    def _normvalue(self, value: Any) -> Any:
        return value


class LocalCache(OrderedDict[_KT, _VT]):
    """Dictionary with a finite number of keys.

    Older items expires first.
    """

    def __init__(self, limit: int | None = None):
        super().__init__()
        self.limit: int | None = limit

    def __setitem__(self, key: _KT, value: _VT) -> None:
        if self.limit is not None:
            if self.limit == 0:
                return
            while len(self) >= self.limit:
                self.popitem(last=False)
        super().__setitem__(key, value)


class LocalWeakReferencedCache(weakref.WeakKeyDictionary[_KT, _VT | None]):
    """
    A weakref.WeakKeyDictionary implementation that uses LocalCache as its
    underlying data structure, making it ordered and capable of being size-limited.

    Useful for memoization, while avoiding keeping received
    arguments in memory only because of the cached references.

    Note: like LocalCache and unlike weakref.WeakKeyDictionary,
    it cannot be instantiated with an initial dictionary.
    """

    def __init__(self, limit: int | None = None):
        super().__init__()
        self.data: LocalCache[_KT, _VT] = LocalCache(limit=limit)

    def __setitem__(self, key: _KT, value: _VT | None) -> None:
        # if raised, key is not weak-referenceable, skip caching
        with contextlib.suppress(TypeError):
            super().__setitem__(key, value)

    def __getitem__(self, key: _KT) -> _VT | None:
        try:
            return cast("_VT", super().__getitem__(key))
        except (TypeError, KeyError):
            return None  # key is either not weak-referenceable or not cached


class SequenceExclude:
    """Object to test if an item is NOT within some sequence."""

    def __init__(self, seq: Container[Any]):
        self.seq: Container[Any] = seq

    def __contains__(self, item: Any) -> bool:
        return item not in self.seq



from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from scrapy.spidermiddlewares.base import BaseSpiderMiddleware

if TYPE_CHECKING:
    from scrapy.crawler import Crawler
    from scrapy.http import Request, Response


logger = logging.getLogger(__name__)


class MetaCopyDetectionMiddleware(BaseSpiderMiddleware):
    """Warn when a spider yields a request with internal meta keys that should
    not be copied from response.meta.

    Each warning is emitted at most once per crawl.
    """

    _INTERNAL_KEYS: frozenset[str] = frozenset(
        {
            "_auth_proxy",
            "_dont_cache",
            "_scheme_proxy",
            "download_latency",
            "redirect_reasons",
            "redirect_times",
            "redirect_ttl",
            "redirect_urls",
            "retry_times",
        }
    )

    def __init__(self, crawler: Crawler) -> None:
        super().__init__(crawler)
        skip = frozenset(crawler.settings.getlist("META_COPY_WARN_SKIP_KEYS", []))
        self._keys: frozenset[str] = self._INTERNAL_KEYS - skip
        self._warned: bool = False

    def get_processed_request(
        self, request: Request, response: Response | None
    ) -> Request | None:
        if response is None:
            return request

        if not self._warned:
            found = self._keys & request.meta.keys()
            if found:
                spider_name = type(self.crawler.spider).__name__
                logger.warning(
                    f"{spider_name} yielded a request containing internal "
                    f"meta keys that were likely copied from response.meta "
                    f"and should not be forwarded to new requests: "
                    f"{sorted(found)}. See the MetaCopyDetectionMiddleware "
                    f"documentation for more information. Source response: "
                    f"{response}, target request: {request}",
                    extra={"spider": self.crawler.spider},
                )
                self._warned = True

        return request



from __future__ import annotations

import asyncio
import logging
import subprocess
import sys
from typing import TYPE_CHECKING, Any
from unittest.mock import Mock

import pytest

from scrapy import signals
from scrapy.core.engine import ExecutionEngine, _Slot
from scrapy.core.scheduler import BaseScheduler
from scrapy.exceptions import CloseSpider, IgnoreRequest
from scrapy.http import Request
from scrapy.spiders import Spider
from scrapy.utils.defer import _schedule_coro, deferred_from_coro
from scrapy.utils.misc import build_from_crawler
from scrapy.utils.spider import DefaultSpider
from scrapy.utils.test import get_crawler
from tests.utils.bases.engine import TestEngineBase
from tests.utils.decorators import coroutine_test, inline_callbacks_test
from tests.utils.engine import (
    AttrsItemsSpider,
    CrawlerRun,
    DataClassItemsSpider,
    DictItemsSpider,
    MySpider,
)

if TYPE_CHECKING:
    from collections.abc import AsyncIterator, Generator

    from twisted.internet.defer import Deferred

    from tests.mockserver.http import MockServer


class DupeFilterSpider(MySpider):
    async def start(self):
        for url in self.start_urls:
            yield Request(url)  # no dont_filter=True


class ItemZeroDivisionErrorSpider(MySpider):
    custom_settings = {
        "ITEM_PIPELINES": {
            "tests.pipelines.ProcessWithZeroDivisionErrorPipeline": 300,
        }
    }


class ChangeCloseReasonSpider(MySpider):
    @classmethod
    def from_crawler(cls, crawler, *args, **kwargs):
        spider = cls(*args, **kwargs)
        spider._set_crawler(crawler)
        crawler.signals.connect(spider.spider_idle, signals.spider_idle)
        return spider

    def spider_idle(self):
        raise CloseSpider(reason="custom_reason")


class TestEngine(TestEngineBase):
    @coroutine_test
    async def test_crawler(self, mockserver: MockServer) -> None:
        for spider in (
            MySpider,
            DictItemsSpider,
            AttrsItemsSpider,
            DataClassItemsSpider,
        ):
            run = CrawlerRun(spider)
            await run.run(mockserver)
            self._assert_visited_urls(run)
            self._assert_scheduled_requests(run, count=9)
            self._assert_downloaded_responses(run, count=9)
            self._assert_scraped_items(run)
            self._assert_signals_caught(run)
            self._assert_headers_received(run)
            self._assert_bytes_received(run)

    @coroutine_test
    async def test_crawler_dupefilter(self, mockserver: MockServer) -> None:
        run = CrawlerRun(DupeFilterSpider)
        await run.run(mockserver)
        self._assert_scheduled_requests(run, count=8)
        self._assert_dropped_requests(run)

    @coroutine_test
    async def test_crawler_itemerror(self, mockserver: MockServer) -> None:
        run = CrawlerRun(ItemZeroDivisionErrorSpider)
        await run.run(mockserver)
        self._assert_items_error(run)

    @coroutine_test
    async def test_crawler_change_close_reason_on_idle(
        self, mockserver: MockServer
    ) -> None:
        run = CrawlerRun(ChangeCloseReasonSpider)
        await run.run(mockserver)
        assert {
            "spider": run.crawler.spider,
            "reason": "custom_reason",
        } == run.signals_caught[signals.spider_closed]

    @coroutine_test
    async def test_close_downloader(self):
        e = ExecutionEngine(get_crawler(MySpider), lambda _: None)
        await e.close_async()

    def test_close_without_downloader(self):
        class CustomException(Exception):
            pass

        class BadDownloader:
            def __init__(self, crawler):
                raise CustomException

        with pytest.raises(CustomException):
            ExecutionEngine(
                get_crawler(MySpider, {"DOWNLOADER": BadDownloader}), lambda _: None
            )

    @inline_callbacks_test
    def test_start_already_running_exception(self):
        crawler = get_crawler(DefaultSpider)
        crawler.spider = crawler._create_spider()
        e = ExecutionEngine(crawler, lambda _: None)
        crawler.engine = e
        yield deferred_from_coro(e.open_spider_async())
        _schedule_coro(e.start_async())
        with pytest.raises(RuntimeError, match="Engine already running"):
            yield deferred_from_coro(e.start_async())
        yield deferred_from_coro(e.stop_async())

    @pytest.mark.only_asyncio
    @coroutine_test
    async def test_start_already_running_exception_asyncio(self):
        crawler = get_crawler(DefaultSpider)
        crawler.spider = crawler._create_spider()
        e = ExecutionEngine(crawler, lambda _: None)
        crawler.engine = e
        await e.open_spider_async()
        with pytest.raises(RuntimeError, match="Engine already running"):
            await asyncio.gather(e.start_async(), e.start_async())
        await e.stop_async()

    @coroutine_test
    async def test_start_request_processing_exception(
        self, caplog: pytest.LogCaptureFixture
    ) -> None:
        class BadRequestFingerprinter:
            def fingerprint(self, request):
                raise ValueError  # to make Scheduler.enqueue_request() fail

        class SimpleSpider(Spider):
            name = "simple"

            async def start(self):
                yield Request("data:,")

        crawler = get_crawler(
            SimpleSpider, {"REQUEST_FINGERPRINTER_CLASS": BadRequestFingerprinter}
        )
        with caplog.at_level(logging.DEBUG):
            await crawler.crawl_async()
        assert "Error while processing requests from start()" in caplog.text
        assert "Spider closed (shutdown)" in caplog.text

    def test_short_timeout(self):
        args = (
            sys.executable,
            "-m",
            "scrapy.cmdline",
            "fetch",
            "-s",
            "CLOSESPIDER_TIMEOUT=0.001",
            "-s",
            "LOG_LEVEL=DEBUG",
            "http://toscrape.com",
        )
        p = subprocess.Popen(
            args,
            stdout=subprocess.DEVNULL,
            stderr=subprocess.PIPE,
        )

        try:
            _, stderr = p.communicate(timeout=15)
        except subprocess.TimeoutExpired:
            p.kill()
            p.communicate()
            pytest.fail("Command took too much time to complete")

        stderr_str = stderr.decode("utf-8")
        assert "AttributeError" not in stderr_str, stderr_str
        assert "AssertionError" not in stderr_str, stderr_str


@coroutine_test
async def test_request_scheduled_signal():
    class TestScheduler(BaseScheduler):
        def __init__(self) -> None:
            self.enqueued: list[Request] = []

        def enqueue_request(self, request: Request) -> bool:
            self.enqueued.append(request)
            return True

    def signal_handler(request: Request, spider: Spider) -> None:
        if "drop" in request.url:
            raise IgnoreRequest

    crawler = get_crawler(MySpider)
    engine = ExecutionEngine(crawler, lambda _: None)
    scheduler = build_from_crawler(TestScheduler, crawler)

    async def start() -> AsyncIterator[Any]:
        return
        yield

    engine._start = start()
    engine._slot = _Slot(False, Mock(), scheduler)
    crawler.signals.connect(signal_handler, signals.request_scheduled)
    keep_request = Request("https://keep.example")
    engine._schedule_request(keep_request)
    drop_request = Request("https://drop.example")
    engine._schedule_request(drop_request)
    assert scheduler.enqueued == [keep_request], (
        f"{scheduler.enqueued!r} != [{keep_request!r}]"
    )
    crawler.signals.disconnect(signal_handler, signals.request_scheduled)


class ClosingPipeline:
    def open_spider(self):
        raise CloseSpider("pipeline_reason")


class TestCloseSpiderOnStartup:
    @coroutine_test
    async def test_pipeline(self, caplog: pytest.LogCaptureFixture) -> None:
        closed: list[str] = []

        def spider_closed(reason: str) -> None:
            closed.append(reason)

        crawler = get_crawler(DefaultSpider, {"ITEM_PIPELINES": {ClosingPipeline: 1}})
        crawler.signals.connect(spider_closed, signals.spider_closed)
        with caplog.at_level(logging.INFO):
            await crawler.crawl_async()
        assert crawler.stats.get_value("finish_reason") == "pipeline_reason"
        assert closed == ["pipeline_reason"]
        assert "Traceback" not in caplog.text

    @coroutine_test
    async def test_spider_opened(self) -> None:
        def spider_opened(spider: Spider) -> None:
            raise CloseSpider("signal_reason")

        crawler = get_crawler(DefaultSpider)
        crawler.signals.connect(spider_opened, signals.spider_opened)
        await crawler.crawl_async()
        assert crawler.stats.get_value("finish_reason") == "signal_reason"

    @coroutine_test
    async def test_startup_wins_over_spider_opened(self) -> None:
        def spider_opened(spider: Spider) -> None:
            raise CloseSpider("signal_reason")

        crawler = get_crawler(DefaultSpider, {"ITEM_PIPELINES": {ClosingPipeline: 1}})
        crawler.signals.connect(spider_opened, signals.spider_opened)
        await crawler.crawl_async()
        assert crawler.stats.get_value("finish_reason") == "pipeline_reason"

    @inline_callbacks_test
    def test_deferred_crawl(self) -> Generator[Deferred[Any], Any, None]:
        crawler = get_crawler(DefaultSpider, {"ITEM_PIPELINES": {ClosingPipeline: 1}})
        yield crawler.crawl()
        assert crawler.stats.get_value("finish_reason") == "pipeline_reason"



from __future__ import annotations

from asyncio import sleep
from typing import TYPE_CHECKING, Any

import pytest

from scrapy import Spider, signals
from scrapy.utils.asyncio import sleep as scrapy_sleep
from scrapy.utils.defer import maybe_deferred_to_future
from scrapy.utils.test import get_crawler

from .utils import twisted_sleep
from .utils.decorators import coroutine_test

if TYPE_CHECKING:
    from collections.abc import AsyncIterator, Callable

SLEEP_SECONDS = 0.1

ITEM_A = {"id": "a"}
ITEM_B = {"id": "b"}


async def _test_spider(
    spider: type[Spider],
    expected_items: list[Any] | None = None,
    settings: dict[str, Any] | None = None,
    *,
    ordered: bool = True,
) -> None:
    actual_items: list[Any] = []
    expected_items = [] if expected_items is None else expected_items

    def track_item(item: Any, response: Any, spider: Spider) -> None:
        actual_items.append(item)

    crawler = get_crawler(spider, settings)
    crawler.signals.connect(track_item, signals.item_scraped)
    await crawler.crawl_async()
    assert crawler.stats.get_value("finish_reason") == "finished"
    if not ordered:
        actual_items = sorted(actual_items, key=repr)
        expected_items = sorted(expected_items, key=repr)

    assert actual_items == expected_items


@coroutine_test
async def test_start_urls() -> None:
    class TestSpider(Spider):
        name = "test"
        start_urls = ["data:,"]

        async def parse(self, response):
            yield ITEM_A

    await _test_spider(TestSpider, [ITEM_A])


@coroutine_test
async def test_start() -> None:
    class TestSpider(Spider):
        name = "test"

        async def start(self):
            yield ITEM_A

    await _test_spider(TestSpider, [ITEM_A])


@coroutine_test
async def test_start_subclass() -> None:
    class BaseSpider(Spider):
        async def start(self):
            yield ITEM_A

    class TestSpider(BaseSpider):
        name = "test"

    await _test_spider(TestSpider, [ITEM_A])


async def _test_start(
    start_: Callable[[Any], AsyncIterator[Any]],
    expected_items: list[Any] | None = None,
) -> None:
    class TestSpider(Spider):
        name = "test"
        start = start_

    await _test_spider(TestSpider, expected_items)


@pytest.mark.only_asyncio
@coroutine_test
async def test_asyncio_delayed() -> None:
    async def start(spider):
        await sleep(SLEEP_SECONDS)
        yield ITEM_A

    await _test_start(start, [ITEM_A])


@pytest.mark.requires_reactor  # needs a reactor for twisted_sleep()
@coroutine_test
async def test_twisted_delayed() -> None:
    async def start(spider):
        await maybe_deferred_to_future(twisted_sleep(SLEEP_SECONDS))
        yield ITEM_A

    await _test_start(start, [ITEM_A])


@coroutine_test
async def test_slow_pipeline() -> None:
    class SlowPipeline:
        async def process_item(self, item):
            await scrapy_sleep(SLEEP_SECONDS)
            return item

    class TestSpider(Spider):
        name = "test"

        async def start(self):
            yield ITEM_A
            yield ITEM_B

    await _test_spider(
        TestSpider,
        [ITEM_A, ITEM_B],
        {"ITEM_PIPELINES": {SlowPipeline: 0}},
        ordered=False,
    )



from __future__ import annotations

from datetime import datetime, timezone
from typing import TYPE_CHECKING

import pytest

from scrapy.exceptions import NotConfigured
from scrapy.extensions.spiderstate import SpiderState
from scrapy.spiders import Spider
from scrapy.utils.misc import build_from_crawler
from scrapy.utils.test import get_crawler

if TYPE_CHECKING:
    from pathlib import Path


def test_store_load(tmp_path: Path) -> None:
    jobdir = str(tmp_path)

    spider = Spider(name="default")
    dt = datetime.now(tz=timezone.utc)

    ss = SpiderState(jobdir)
    ss.spider_opened(spider)
    assert hasattr(spider, "state")
    spider.state["one"] = 1
    spider.state["dt"] = dt
    ss.spider_closed(spider)

    spider2 = Spider(name="default")
    ss2 = SpiderState(jobdir)
    ss2.spider_opened(spider2)
    assert hasattr(spider2, "state")
    assert spider2.state == {"one": 1, "dt": dt}
    ss2.spider_closed(spider2)


def test_state_attribute() -> None:
    # state attribute must be present if jobdir is not set, to provide a
    # consistent interface
    spider = Spider(name="default")
    ss = SpiderState()
    ss.spider_opened(spider)
    assert hasattr(spider, "state")
    assert spider.state == {}
    ss.spider_closed(spider)


def test_not_configured() -> None:
    crawler = get_crawler(Spider)
    with pytest.raises(NotConfigured):
        build_from_crawler(SpiderState, crawler)



from __future__ import annotations

import inspect
import warnings
from unittest import mock

import pytest

from scrapy.exceptions import ScrapyDeprecationWarning
from scrapy.utils.deprecate import attribute, create_deprecated_class, update_classpath


class MyWarning(UserWarning):
    pass


class SomeBaseClass:
    pass


class NewName(SomeBaseClass):
    pass


class TestWarnWhenSubclassed:
    def test_no_warning_on_definition(self):
        with warnings.catch_warnings():
            warnings.simplefilter("error", category=ScrapyDeprecationWarning)
            create_deprecated_class("Deprecated", NewName)

    def test_subclassing_warning_message(self):
        msg = (
            r"tests\.test_utils_deprecate\.UserClass inherits from "
            r"deprecated class tests\.test_utils_deprecate\.Deprecated, "
            r"please inherit from tests\.test_utils_deprecate\.NewName."
            r" \(warning only on first subclass, there may be others\)"
        )
        Deprecated = create_deprecated_class(
            "Deprecated", NewName, warn_category=MyWarning
        )
        with pytest.warns(MyWarning, match=msg) as w:

            class UserClass(Deprecated):  # type: ignore[misc, valid-type]
                pass

        assert w[0].lineno == inspect.getsourcelines(UserClass)[1]

    def test_custom_class_paths(self):
        Deprecated = create_deprecated_class(
            "Deprecated",
            NewName,
            new_class_path="foo.NewClass",
            old_class_path="bar.OldClass",
            warn_category=MyWarning,
        )

        with pytest.warns(
            MyWarning,
            match=r"UserClass inherits from deprecated class bar\.OldClass, please inherit from foo\.NewClass",
        ):

            class UserClass(Deprecated):  # type: ignore[misc, valid-type]
                pass

        with pytest.warns(
            MyWarning,
            match=r"bar\.OldClass is deprecated, instantiate foo\.NewClass instead",
        ):
            _ = Deprecated()

    def test_subclassing_warns_only_on_direct_children(self):
        Deprecated = create_deprecated_class(
            "Deprecated", NewName, warn_once=False, warn_category=MyWarning
        )

        with pytest.warns(
            MyWarning,
            match="UserClass inherits from deprecated class",
        ):

            class UserClass(Deprecated):  # type: ignore[misc, valid-type]
                pass

        with warnings.catch_warnings():
            warnings.simplefilter("error", MyWarning)

            class NoWarnOnMe(UserClass):
                pass

    def test_subclassing_warns_once_by_default(self):
        Deprecated = create_deprecated_class(
            "Deprecated", NewName, warn_category=MyWarning
        )

        with pytest.warns(
            MyWarning,
            match="UserClass inherits from deprecated class",
        ):

            class UserClass(Deprecated):  # type: ignore[misc, valid-type]
                pass

        with warnings.catch_warnings():
            warnings.simplefilter("error", MyWarning)

            class FooClass(Deprecated):  # type: ignore[misc, valid-type]
                pass

            class BarClass(Deprecated):  # type: ignore[misc, valid-type]
                pass

    def test_warning_on_instance(self):
        Deprecated = create_deprecated_class(
            "Deprecated", NewName, warn_category=MyWarning
        )

        with pytest.warns(
            MyWarning,
            match=r"tests\.test_utils_deprecate\.Deprecated is deprecated, "
            r"instantiate tests\.test_utils_deprecate\.NewName instead\.",
        ) as w:
            _, lineno = Deprecated(), inspect.getlineno(inspect.currentframe())  # type: ignore[arg-type]
        assert len(w) == 1
        assert w[0].lineno == lineno

        # ignore subclassing warnings
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", MyWarning)

            class UserClass(Deprecated):  # type: ignore[misc, valid-type]
                pass

        with warnings.catch_warnings():
            warnings.simplefilter("error", MyWarning)
            UserClass()  # subclass instances don't warn

    def test_warning_auto_message(self):
        Deprecated = create_deprecated_class("Deprecated", NewName)
        with pytest.warns(
            ScrapyDeprecationWarning,
            match=r"UserClass2 inherits from deprecated class tests\.test_utils_deprecate\.Deprecated, please inherit from tests\.test_utils_deprecate\.NewName",
        ):

            class UserClass2(Deprecated):  # type: ignore[misc, valid-type]
                pass

    def test_issubclass(self):
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", ScrapyDeprecationWarning)
            DeprecatedName = create_deprecated_class("DeprecatedName", NewName)

            class UpdatedUserClass1(NewName):
                pass

            class UpdatedUserClass1a(NewName):
                pass

            class OutdatedUserClass1(DeprecatedName):  # type: ignore[misc, valid-type]
                pass

            class OutdatedUserClass1a(DeprecatedName):  # type: ignore[misc, valid-type]
                pass

            class UnrelatedClass:
                pass

        assert issubclass(UpdatedUserClass1, NewName)
        assert issubclass(UpdatedUserClass1a, NewName)
        assert issubclass(UpdatedUserClass1, DeprecatedName)
        assert issubclass(UpdatedUserClass1a, DeprecatedName)
        assert issubclass(OutdatedUserClass1, DeprecatedName)
        assert not issubclass(UnrelatedClass, DeprecatedName)
        assert not issubclass(OutdatedUserClass1, OutdatedUserClass1a)
        assert not issubclass(OutdatedUserClass1a, OutdatedUserClass1)

        with pytest.raises(TypeError):
            issubclass(object(), DeprecatedName)  # type: ignore[arg-type]

    def test_isinstance(self):
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", ScrapyDeprecationWarning)
            DeprecatedName = create_deprecated_class("DeprecatedName", NewName)

            class UpdatedUserClass2(NewName):
                pass

            class UpdatedUserClass2a(NewName):
                pass

            class OutdatedUserClass2(DeprecatedName):  # type: ignore[misc, valid-type]
                pass

            class OutdatedUserClass2a(DeprecatedName):  # type: ignore[misc, valid-type]
                pass

            class UnrelatedClass:
                pass

        assert isinstance(UpdatedUserClass2(), NewName)
        assert isinstance(UpdatedUserClass2a(), NewName)
        assert isinstance(UpdatedUserClass2(), DeprecatedName)
        assert isinstance(UpdatedUserClass2a(), DeprecatedName)
        assert isinstance(OutdatedUserClass2(), DeprecatedName)
        assert isinstance(OutdatedUserClass2a(), DeprecatedName)
        assert not isinstance(OutdatedUserClass2a(), OutdatedUserClass2)
        assert not isinstance(OutdatedUserClass2(), OutdatedUserClass2a)
        assert not isinstance(UnrelatedClass(), DeprecatedName)

    def test_clsdict(self):
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", ScrapyDeprecationWarning)
            Deprecated = create_deprecated_class("Deprecated", NewName, {"foo": "bar"})

        assert Deprecated.foo == "bar"  # type: ignore[attr-defined]

    def test_deprecate_a_class_with_custom_metaclass(self):
        Meta1 = type("Meta1", (type,), {})
        New = Meta1("New", (), {})
        create_deprecated_class("Deprecated", New)

    def test_deprecate_subclass_of_deprecated_class(self):
        with warnings.catch_warnings():
            warnings.simplefilter("error", MyWarning)
            Deprecated = create_deprecated_class(
                "Deprecated", NewName, warn_category=MyWarning
            )
            AlsoDeprecated = create_deprecated_class(
                "AlsoDeprecated",
                Deprecated,
                new_class_path="foo.Bar",
                warn_category=MyWarning,
            )

        with pytest.warns(
            MyWarning,
            match=r"AlsoDeprecated is deprecated, instantiate foo\.Bar instead",
        ):
            AlsoDeprecated()

        with pytest.warns(
            MyWarning,
            match=r"UserClass inherits from deprecated class tests\.test_utils_deprecate\.AlsoDeprecated, please inherit from foo\.Bar",
        ):

            class UserClass(AlsoDeprecated):  # type: ignore[misc, valid-type]
                pass

    def test_inspect_stack(self):
        with (
            mock.patch("inspect.stack", side_effect=IndexError),
            pytest.warns(UserWarning, match="Error detecting parent module"),
        ):
            create_deprecated_class("DeprecatedName", NewName)

    def test_unknown_parent_module(self):
        with mock.patch("inspect.getmodule", return_value=None):
            cls = create_deprecated_class("DeprecatedName", NewName)
        assert cls.__module__ == "scrapy.utils.deprecate"


@mock.patch(
    "scrapy.utils.deprecate.DEPRECATION_RULES",
    [
        ("scrapy.contrib.pipeline.", "scrapy.pipelines."),
        ("scrapy.contrib.", "scrapy.extensions."),
    ],
)
class TestUpdateClassPath:
    def test_old_path_gets_fixed(self):
        with pytest.warns(
            ScrapyDeprecationWarning,
            match="`scrapy.contrib.debug.Debug` class is deprecated, use `scrapy.extensions.debug.Debug` instead",
        ):
            output = update_classpath("scrapy.contrib.debug.Debug")
        assert output == "scrapy.extensions.debug.Debug"

    def test_sorted_replacement(self):
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", ScrapyDeprecationWarning)
            output = update_classpath("scrapy.contrib.pipeline.Pipeline")
        assert output == "scrapy.pipelines.Pipeline"

    def test_unmatched_path_stays_the_same(self):
        with warnings.catch_warnings():
            warnings.simplefilter("error", ScrapyDeprecationWarning)
            output = update_classpath("scrapy.unmatched.Path")
        assert output == "scrapy.unmatched.Path"

    def test_returns_nonstring(self):
        for notastring in [None, True, [1, 2, 3], object()]:
            assert update_classpath(notastring) == notastring


class TestAttribute:
    class MyClass:
        pass

    def test_default_version(self):
        with pytest.warns(
            ScrapyDeprecationWarning,
            match=r"MyClass\.old attribute is deprecated and will be no longer "
            r"supported in Scrapy 0\.12, use MyClass\.new attribute instead",
        ):
            attribute(self.MyClass(), "old", "new")

    def test_custom_version(self):
        with pytest.warns(ScrapyDeprecationWarning, match="in Scrapy 3.0"):
            attribute(self.MyClass(), "old", "new", "3.0")



from scrapy.signalmanager import SignalManager


def test_disconnect_all() -> None:
    signal = object()
    sender = object()
    sm = SignalManager(sender)

    calls = []

    def handler():
        calls.append(1)

    sm.connect(handler, signal)
    sm.send_catch_log(signal)
    assert calls == [1]

    sm.disconnect_all(signal)
    sm.send_catch_log(signal)
    assert calls == [1]  # handler no longer called after disconnect_all



from __future__ import annotations

from collections.abc import Mapping
from typing import TYPE_CHECKING, Any, TypeAlias, cast

from w3lib.http import headers_dict_to_raw

from scrapy.utils.datatypes import CaseInsensitiveDict
from scrapy.utils.python import to_unicode

if TYPE_CHECKING:
    from collections.abc import Iterable

    # typing.Self requires Python 3.11
    from typing_extensions import Self


_RawValue: TypeAlias = bytes | str | int


class Headers(dict):  # type: ignore[type-arg]
    """Case insensitive http headers dictionary"""

    def __init__(
        self,
        seq: Mapping[str, Any]
        | Mapping[bytes, Any]
        | Iterable[tuple[str | bytes, Any]]
        | None = None,
        encoding: str = "utf-8",
    ):
        self.encoding: str = encoding
        super().__init__()
        if seq:
            self.update(seq)

    def __setitem__(self, key: str | bytes, value: Any) -> None:
        dict.__setitem__(self, self.normkey(key), self.normvalue(value))

    def __delitem__(self, key: str | bytes) -> None:
        dict.__delitem__(self, self.normkey(key))

    def __contains__(self, key: str | bytes) -> bool:  # type: ignore[override]
        return dict.__contains__(self, self.normkey(key))

    has_key = __contains__

    def setdefault(self, key: str | bytes, def_val: Any = None) -> Any:
        return dict.setdefault(self, self.normkey(key), self.normvalue(def_val))

    @classmethod
    def fromkeys(  # type: ignore[override]
        cls, keys: Iterable[str | bytes], value: Any = None
    ) -> Self:
        return cls((k, value) for k in keys)

    def pop(self, key: str | bytes, *args: Any) -> Any:
        return dict.pop(self, self.normkey(key), *args)

    def update(  # type: ignore[override]
        self,
        seq: Mapping[str, Any]
        | Mapping[bytes, Any]
        | Iterable[tuple[str | bytes, Any]],
    ) -> None:
        seq = seq.items() if isinstance(seq, Mapping) else seq
        iseq: dict[bytes, list[bytes]] = {}
        for k, v in seq:
            iseq.setdefault(self.normkey(k), []).extend(self.normvalue(v))
        dict.update(self, iseq)

    def normkey(self, key: str | bytes) -> bytes:
        """Normalize key to bytes"""
        return self._tobytes(key.title())

    def normvalue(self, value: _RawValue | Iterable[_RawValue]) -> list[bytes]:
        """Normalize values to bytes"""
        _value: Iterable[_RawValue]
        if value is None:
            _value = []
        elif isinstance(value, (str, bytes)):
            _value = [value]
        elif hasattr(value, "__iter__"):
            _value = value
        else:
            _value = [value]

        return [self._tobytes(x) for x in _value]

    def _tobytes(self, x: _RawValue) -> bytes:
        if isinstance(x, bytes):
            return x
        if isinstance(x, str):
            return x.encode(self.encoding)
        if isinstance(x, int):
            return str(x).encode(self.encoding)
        raise TypeError(f"Unsupported value type: {type(x)}")

    def __getitem__(self, key: str | bytes) -> bytes | None:
        try:
            return cast("list[bytes]", dict.__getitem__(self, self.normkey(key)))[-1]
        except IndexError:
            return None

    def get(self, key: str | bytes, def_val: Any = None) -> bytes | None:
        try:
            return cast(
                "list[bytes]",
                dict.get(self, self.normkey(key), self.normvalue(def_val)),
            )[-1]
        except IndexError:
            return None

    def getlist(self, key: str | bytes, def_val: Any = None) -> list[bytes]:
        try:
            return cast("list[bytes]", dict.__getitem__(self, self.normkey(key)))
        except KeyError:
            if def_val is not None:
                return self.normvalue(def_val)
            return []

    def setlist(self, key: str | bytes, list_: Iterable[_RawValue]) -> None:
        self[key] = list_

    def setlistdefault(
        self, key: str | bytes, default_list: Iterable[_RawValue] = ()
    ) -> Any:
        return self.setdefault(key, default_list)

    def appendlist(self, key: str | bytes, value: Iterable[_RawValue]) -> None:
        lst = self.getlist(key)
        lst.extend(self.normvalue(value))
        self[key] = lst

    def items(self) -> Iterable[tuple[bytes, list[bytes]]]:  # type: ignore[override]
        return ((k, self.getlist(k)) for k in self.keys())

    def values(self) -> list[bytes | None]:  # type: ignore[override]
        return [
            self[k]
            for k in self.keys()  # pylint: disable=consider-using-dict-items
        ]

    def to_string(self) -> bytes:
        return headers_dict_to_raw(self)

    def to_unicode_dict(self) -> CaseInsensitiveDict:
        """Return headers as a CaseInsensitiveDict with str keys
        and str values. Multiple values are joined with ','.
        """
        return CaseInsensitiveDict(
            (
                to_unicode(key, encoding=self.encoding),
                to_unicode(b",".join(value), encoding=self.encoding),
            )
            for key, value in self.items()
        )

    def to_tuple_list(self) -> list[tuple[str, str]]:
        """Return headers as a list of ``(key, value)`` tuples.

        Multiple values are represented as multiple tuples with the same key.
        """
        return [
            (key.decode(self.encoding), value.decode(self.encoding))
            for key, values in self.items()
            for value in values
        ]

    def __copy__(self) -> Self:
        return self.__class__(self)

    copy = __copy__



from __future__ import annotations

import datetime
import json
import logging
from typing import TYPE_CHECKING, Any

import pytest

from scrapy.exceptions import NotConfigured, ScrapyDeprecationWarning
from scrapy.extensions.periodic_log import PeriodicLog
from scrapy.utils.misc import build_from_crawler
from scrapy.utils.test import get_crawler

from .spiders import MetaSpider
from .utils.decorators import coroutine_test

if TYPE_CHECKING:
    from collections.abc import Callable


stats_dump_1 = {
    "log_count/INFO": 10,
    "log_count/WARNING": 1,
    "start_time": datetime.datetime(2023, 6, 16, 8, 59, 18, 993170),
    "scheduler/enqueued/memory": 190,
    "scheduler/enqueued": 190,
    "scheduler/dequeued/memory": 166,
    "scheduler/dequeued": 166,
    "downloader/request_count": 166,
    "downloader/request_method_count/GET": 166,
    "downloader/request_bytes": 56803,
    "downloader/response_count": 150,
    "downloader/response_status_count/200": 150,
    "downloader/response_bytes": 595698,
    "httpcompression/response_bytes": 3186068,
    "httpcompression/response_count": 150,
    "response_received_count": 150,
    "request_depth_max": 9,
    "dupefilter/filtered": 180,
    "item_scraped_count": 140,
}
stats_dump_2 = {
    "log_count/INFO": 12,
    "log_count/WARNING": 1,
    "start_time": datetime.datetime(2023, 6, 16, 8, 59, 18, 993170),
    "scheduler/enqueued/memory": 337,
    "scheduler/enqueued": 337,
    "scheduler/dequeued/memory": 280,
    "scheduler/dequeued": 280,
    "downloader/request_count": 280,
    "downloader/request_method_count/GET": 280,
    "downloader/request_bytes": 95754,
    "downloader/response_count": 264,
    "downloader/response_status_count/200": 264,
    "downloader/response_bytes": 1046274,
    "httpcompression/response_bytes": 5614484,
    "httpcompression/response_count": 264,
    "response_received_count": 264,
    "request_depth_max": 16,
    "dupefilter/filtered": 320,
    "item_scraped_count": 248,
}


class CustomPeriodicLog(PeriodicLog):
    def set_a(self) -> None:
        self.stats._stats = stats_dump_1

    def set_b(self) -> None:
        self.stats._stats = stats_dump_2


def extension(settings: dict[str, Any] | None = None) -> CustomPeriodicLog:
    crawler = get_crawler(MetaSpider, settings)
    return build_from_crawler(CustomPeriodicLog, crawler)


def test_extension_enabled() -> None:
    # Expected that settings for this extension loaded successfully

    # "PERIODIC_LOG_STATS": True -> set to {"enabled": True}
    # due to TypeError exception from settings.getdict
    assert extension({"PERIODIC_LOG_STATS": True, "LOGSTATS_INTERVAL": 60})

    # "PERIODIC_LOG_STATS": "True" -> set to {"enabled": True}
    # due to JSONDecodeError(ValueError) exception from settings.getdict
    assert extension({"PERIODIC_LOG_STATS": "True", "LOGSTATS_INTERVAL": 60})

    # The ame for PERIODIC_LOG_DELTA:
    assert extension({"PERIODIC_LOG_DELTA": True, "LOGSTATS_INTERVAL": 60})
    assert extension({"PERIODIC_LOG_DELTA": "True", "LOGSTATS_INTERVAL": 60})


def test_no_interval() -> None:
    with pytest.raises(NotConfigured):
        extension({"PERIODIC_LOG_STATS": True, "LOGSTATS_INTERVAL": 0})


def test_nothing_enabled() -> None:
    with pytest.raises(NotConfigured):
        extension({"LOGSTATS_INTERVAL": 60})


@coroutine_test
async def test_log_delta() -> None:
    def emulate(
        settings: dict[str, Any] | None = None,
    ) -> tuple[PeriodicLog, dict[str, Any], dict[str, Any]]:
        spider = MetaSpider()
        ext = extension(settings)
        ext.spider_opened(spider)
        ext.set_a()
        a = ext.log_delta()
        ext.set_b()
        b = ext.log_delta()
        ext.spider_closed(spider, reason="finished")
        return ext, a, b

    def check(settings: dict[str, Any], condition: Callable[[str, Any], bool]) -> None:
        ext, a, b = emulate(settings)
        assert list(a["delta"].keys()) == [
            k for k, v in ext.stats._stats.items() if condition(k, v)
        ]
        assert list(b["delta"].keys()) == [
            k for k, v in ext.stats._stats.items() if condition(k, v)
        ]

    # Including all
    check({"PERIODIC_LOG_DELTA": True}, lambda k, v: isinstance(v, (int, float)))

    # include:
    check(
        {"PERIODIC_LOG_DELTA": {"include": ["downloader/"]}},
        lambda k, v: isinstance(v, (int, float)) and "downloader/" in k,
    )

    # include multiple
    check(
        {"PERIODIC_LOG_DELTA": {"include": ["downloader/", "scheduler/"]}},
        lambda k, v: (
            isinstance(v, (int, float)) and ("downloader/" in k or "scheduler/" in k)
        ),
    )

    # exclude
    check(
        {"PERIODIC_LOG_DELTA": {"exclude": ["downloader/"]}},
        lambda k, v: isinstance(v, (int, float)) and "downloader/" not in k,
    )

    # exclude multiple
    check(
        {"PERIODIC_LOG_DELTA": {"exclude": ["downloader/", "scheduler/"]}},
        lambda k, v: (
            isinstance(v, (int, float))
            and ("downloader/" not in k and "scheduler/" not in k)
        ),
    )

    # include exclude combined
    check(
        {"PERIODIC_LOG_DELTA": {"include": ["downloader/"], "exclude": ["bytes"]}},
        lambda k, v: (
            isinstance(v, (int, float)) and ("downloader/" in k and "bytes" not in k)
        ),
    )


@coroutine_test
async def test_log_stats() -> None:
    def emulate(
        settings: dict[str, Any] | None = None,
    ) -> tuple[PeriodicLog, dict[str, Any], dict[str, Any]]:
        spider = MetaSpider()
        ext = extension(settings)
        ext.spider_opened(spider)
        ext.set_a()
        a = ext.log_crawler_stats()
        ext.set_b()
        b = ext.log_crawler_stats()
        ext.spider_closed(spider, reason="finished")
        return ext, a, b

    def check(settings: dict[str, Any], condition: Callable[[str, Any], bool]) -> None:
        ext, a, b = emulate(settings)
        assert list(a["stats"].keys()) == [
            k for k, v in ext.stats._stats.items() if condition(k, v)
        ]
        assert list(b["stats"].keys()) == [
            k for k, v in ext.stats._stats.items() if condition(k, v)
        ]

    # Including all
    check({"PERIODIC_LOG_STATS": True}, lambda k, v: True)

    # include:
    check(
        {"PERIODIC_LOG_STATS": {"include": ["downloader/"]}},
        lambda k, v: "downloader/" in k,
    )

    # include multiple
    check(
        {"PERIODIC_LOG_STATS": {"include": ["downloader/", "scheduler/"]}},
        lambda k, v: "downloader/" in k or "scheduler/" in k,
    )

    # exclude
    check(
        {"PERIODIC_LOG_STATS": {"exclude": ["downloader/"]}},
        lambda k, v: "downloader/" not in k,
    )

    # exclude multiple
    check(
        {"PERIODIC_LOG_STATS": {"exclude": ["downloader/", "scheduler/"]}},
        lambda k, v: "downloader/" not in k and "scheduler/" not in k,
    )

    # include exclude combined
    check(
        {"PERIODIC_LOG_STATS": {"include": ["downloader/"], "exclude": ["bytes"]}},
        lambda k, v: "downloader/" in k and "bytes" not in k,
    )


@coroutine_test
async def test_log_timing(caplog: pytest.LogCaptureFixture) -> None:
    settings = {
        "EXTENSIONS": {"scrapy.extensions.periodic_log.PeriodicLog": 0},
        "PERIODIC_LOG_TIMING_ENABLED": True,
        "LOGSTATS_INTERVAL": 30,
    }
    crawler = get_crawler(MetaSpider, settings)
    with caplog.at_level(logging.INFO, logger="scrapy.extensions.periodic_log"):
        await crawler.crawl_async()

    records = [r for r in caplog.records if r.name == "scrapy.extensions.periodic_log"]
    assert records, "PeriodicLog logged nothing"
    # Only the timing section is enabled, and it is logged on spider close.
    data = json.loads(records[-1].getMessage())
    assert list(data) == ["time"]
    assert data["time"]["log_interval"] == 30
    assert data["time"]["log_interval_real"] >= 0
    assert data["time"]["elapsed"] >= 0
    assert data["time"]["start_time"] <= data["time"]["utcnow"]


def test_multiplier_deprecated() -> None:
    crawler = get_crawler(
        MetaSpider,
        {"PERIODIC_LOG_TIMING_ENABLED": True, "LOGSTATS_INTERVAL": 30},
    )
    crawler._apply_settings()
    ext = build_from_crawler(PeriodicLog, crawler)
    with pytest.warns(ScrapyDeprecationWarning):
        assert ext.multiplier == 2.0
    with pytest.warns(ScrapyDeprecationWarning):
        ext.multiplier = 3.0
    with pytest.warns(ScrapyDeprecationWarning):
        assert ext.multiplier == 3.0



"""
Queues that handle requests
"""

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

import pytest

from scrapy.http import Request
from scrapy.spiders import Spider
from scrapy.squeues import (
    FifoMemoryQueue,
    LifoMemoryQueue,
    MarshalFifoDiskQueue,
    MarshalLifoDiskQueue,
    PickleFifoDiskQueue,
    PickleLifoDiskQueue,
)
from scrapy.utils.misc import build_from_crawler
from scrapy.utils.test import get_crawler

if TYPE_CHECKING:
    import queuelib

    from scrapy.crawler import Crawler


@pytest.fixture
def crawler() -> Crawler:
    return get_crawler(Spider)


class TestRequestQueueBase(ABC):
    @property
    @abstractmethod
    def is_fifo(self) -> bool:
        raise NotImplementedError

    def test_one_element(self, q: queuelib.queue.BaseQueue):
        assert len(q) == 0
        assert q.peek() is None
        assert q.pop() is None
        req = Request("http://www.example.com")
        q.push(req)
        assert len(q) == 1
        result = q.peek()
        assert result is not None
        assert result.url == req.url
        result = q.pop()
        assert result is not None
        assert result.url == req.url
        assert len(q) == 0
        assert q.peek() is None
        assert q.pop() is None

    def test_order(self, q: queuelib.queue.BaseQueue):
        assert len(q) == 0
        assert q.peek() is None
        assert q.pop() is None
        req1 = Request("http://www.example.com/1")
        req2 = Request("http://www.example.com/2")
        req3 = Request("http://www.example.com/3")
        q.push(req1)
        q.push(req2)
        q.push(req3)
        reqs = [req1, req2, req3] if self.is_fifo else [req3, req2, req1]
        for i, req in enumerate(reqs):
            assert len(q) == 3 - i
            result = q.peek()
            assert result is not None
            assert result.url == req.url
            result = q.pop()
            assert result is not None
            assert result.url == req.url
        assert len(q) == 0
        assert q.peek() is None
        assert q.pop() is None


class TestPickleFifoDiskQueueRequest(TestRequestQueueBase):
    is_fifo = True

    @pytest.fixture
    def q(self, crawler, tmp_path):
        queue = build_from_crawler(
            PickleFifoDiskQueue, crawler, key=str(tmp_path / "pickle" / "fifo")
        )
        try:
            yield queue
        finally:
            queue.close()


class TestPickleLifoDiskQueueRequest(TestRequestQueueBase):
    is_fifo = False

    @pytest.fixture
    def q(self, crawler, tmp_path):
        queue = build_from_crawler(
            PickleLifoDiskQueue, crawler, key=str(tmp_path / "pickle" / "lifo")
        )
        try:
            yield queue
        finally:
            queue.close()


class TestMarshalFifoDiskQueueRequest(TestRequestQueueBase):
    is_fifo = True

    @pytest.fixture
    def q(self, crawler, tmp_path):
        queue = build_from_crawler(
            MarshalFifoDiskQueue, crawler, key=str(tmp_path / "marshal" / "fifo")
        )
        try:
            yield queue
        finally:
            queue.close()


class TestMarshalLifoDiskQueueRequest(TestRequestQueueBase):
    is_fifo = False

    @pytest.fixture
    def q(self, crawler, tmp_path):
        queue = build_from_crawler(
            MarshalLifoDiskQueue, crawler, key=str(tmp_path / "marshal" / "lifo")
        )
        try:
            yield queue
        finally:
            queue.close()


class TestFifoMemoryQueueRequest(TestRequestQueueBase):
    is_fifo = True

    @pytest.fixture
    def q(self, crawler):
        return build_from_crawler(FifoMemoryQueue, crawler)


class TestLifoMemoryQueueRequest(TestRequestQueueBase):
    is_fifo = False

    @pytest.fixture
    def q(self, crawler):
        return build_from_crawler(LifoMemoryQueue, crawler)



import datetime
import decimal
import json
from typing import Any

from itemadapter import ItemAdapter, is_item
from twisted.internet import defer

from scrapy.http import Request, Response


class ScrapyJSONEncoder(json.JSONEncoder):
    def default(self, o: Any) -> Any:
        if isinstance(o, set):
            return list(o)
        if isinstance(o, (datetime.datetime, datetime.date, datetime.time)):
            return o.isoformat()
        if isinstance(o, decimal.Decimal):
            return str(o)
        if isinstance(o, defer.Deferred):
            return str(o)
        if isinstance(o, Request):
            return f"<{type(o).__name__} {o.method} {o.url}>"
        if isinstance(o, Response):
            return f"<{type(o).__name__} {o.status} {o.url}>"
        if is_item(o):
            return ItemAdapter(o).asdict()
        return super().default(o)



from __future__ import annotations

from typing import TYPE_CHECKING, Any

import attr
from twisted.internet import defer
from twisted.internet.base import ReactorBase, ThreadedResolver
from twisted.internet.interfaces import (
    IAddress,
    IHostnameResolver,
    IHostResolution,
    IResolutionReceiver,
    IResolverSimple,
)
from zope.interface.declarations import implementer, provider

from scrapy.utils.datatypes import LocalCache

if TYPE_CHECKING:
    from collections.abc import Sequence

    from twisted.internet.defer import Deferred

    # typing.Self requires Python 3.11
    from typing_extensions import Self

    from scrapy.crawler import Crawler

# TODO: cache misses
dnscache: LocalCache[str, Any] = LocalCache(10000)


@implementer(IResolverSimple)
class CachingThreadedResolver(ThreadedResolver):
    """
    Default caching resolver. IPv4 only, supports setting a timeout value for DNS requests.
    """

    def __init__(self, reactor: ReactorBase, cache_size: int, timeout: float):
        super().__init__(reactor)
        dnscache.limit = cache_size
        self.timeout = timeout

    @classmethod
    def from_crawler(cls, crawler: Crawler, reactor: ReactorBase) -> Self:
        if crawler.settings.getbool("DNSCACHE_ENABLED"):
            cache_size = crawler.settings.getint("DNSCACHE_SIZE")
        else:
            cache_size = 0
        return cls(reactor, cache_size, crawler.settings.getfloat("DNS_TIMEOUT"))

    def install_on_reactor(self) -> None:
        self.reactor.installResolver(self)

    def getHostByName(self, name: str, timeout: Sequence[int] = ()) -> Deferred[str]:
        if name in dnscache:
            return defer.succeed(dnscache[name])
        # in Twisted<=16.6, getHostByName() is always called with
        # a default timeout of 60s (actually passed as (1, 3, 11, 45) tuple),
        # so the input argument above is simply overridden
        # to enforce Scrapy's DNS_TIMEOUT setting's value
        # The timeout arg is typed as Sequence[int] but supports floats.
        timeout = (self.timeout,)  # type: ignore[assignment]
        d = super().getHostByName(name, timeout)
        if dnscache.limit:
            d.addCallback(self._cache_result, name)
        return d

    def _cache_result(self, result: Any, name: str) -> Any:
        dnscache[name] = result
        return result


def _address_with_port(address: IAddress, port: int) -> IAddress:
    if getattr(address, "port", port) == port:
        return address
    return attr.evolve(address, port=port)


@implementer(IHostResolution)
class HostResolution:
    def __init__(self, name: str):
        self.name: str = name

    def cancel(self) -> None:  # pragma: no cover
        raise NotImplementedError


@provider(IResolutionReceiver)
class _CachingResolutionReceiver:
    def __init__(self, resolutionReceiver: IResolutionReceiver, hostName: str):
        self.resolutionReceiver: IResolutionReceiver = resolutionReceiver
        self.hostName: str = hostName
        self.addresses: list[IAddress] = []

    def resolutionBegan(self, resolution: IHostResolution) -> None:
        self.resolutionReceiver.resolutionBegan(resolution)
        self.resolution = resolution

    def addressResolved(self, address: IAddress) -> None:
        self.resolutionReceiver.addressResolved(address)
        self.addresses.append(address)

    def resolutionComplete(self) -> None:
        self.resolutionReceiver.resolutionComplete()
        if self.addresses:
            # Name resolution does not depend on the port, so cache entries are
            # kept port-agnostic and the requested port is set on cache hits.
            dnscache[self.hostName] = [
                _address_with_port(address, 0) for address in self.addresses
            ]


@implementer(IHostnameResolver)
class CachingHostnameResolver:
    """
    Experimental caching resolver. Resolves IPv4 and IPv6 addresses,
    does not support setting a timeout value for DNS requests.
    """

    def __init__(self, reactor: ReactorBase, cache_size: int):
        self.reactor: ReactorBase = reactor
        self.original_resolver: IHostnameResolver = reactor.nameResolver
        dnscache.limit = cache_size

    @classmethod
    def from_crawler(cls, crawler: Crawler, reactor: ReactorBase) -> Self:
        if crawler.settings.getbool("DNSCACHE_ENABLED"):
            cache_size = crawler.settings.getint("DNSCACHE_SIZE")
        else:
            cache_size = 0
        return cls(reactor, cache_size)

    def install_on_reactor(self) -> None:
        self.reactor.installNameResolver(self)

    def resolveHostName(
        self,
        resolutionReceiver: IResolutionReceiver,
        hostName: str,
        portNumber: int = 0,
        addressTypes: Sequence[type[IAddress]] | None = None,
        transportSemantics: str = "TCP",
    ) -> IHostResolution:
        try:
            addresses = dnscache[hostName]
        except KeyError:
            return self.original_resolver.resolveHostName(
                _CachingResolutionReceiver(resolutionReceiver, hostName),
                hostName,
                portNumber,
                addressTypes,
                transportSemantics,
            )
        resolutionReceiver.resolutionBegan(HostResolution(hostName))
        for address in addresses:
            resolutionReceiver.addressResolved(_address_with_port(address, portNumber))
        resolutionReceiver.resolutionComplete()
        return resolutionReceiver



from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

from scrapy.http import Request, Response, TextResponse, XmlResponse
from scrapy.spiders import CSVFeedSpider, Spider, XMLFeedSpider
from tests import get_testdata
from tests.spiders import RawResponseSpider
from tests.utils.bases.spider import TestSpiderBase
from tests.utils.crawl import crawl_items
from tests.utils.decorators import coroutine_test

if TYPE_CHECKING:
    from tests.mockserver.http import MockServer


class RawFeedSpider(RawResponseSpider):
    content_type = "text/xml"

    async def start(self):
        yield Request(self.raw_url)


class TestSpider(TestSpiderBase):
    spider_class = Spider


class TestXMLFeedSpider(TestSpiderBase):
    spider_class = XMLFeedSpider

    def test_register_namespace(self):
        body = b"""<?xml version="1.0" encoding="UTF-8"?>
        <urlset xmlns:x="http://www.google.com/schemas/sitemap/0.84"
                xmlns:y="http://www.example.com/schemas/extras/1.0">
        <url><x:loc>http://www.example.com/Special-Offers.html</x:loc><y:updated>2009-08-16</y:updated>
            <other value="bar" y:custom="fuu"/>
        </url>
        <url><loc>http://www.example.com/</loc><y:updated>2009-08-16</y:updated><other value="foo"/></url>
        </urlset>"""
        response = XmlResponse(url="http://example.com/sitemap.xml", body=body)

        class _XMLSpider(self.spider_class):  # type: ignore[name-defined,misc]
            itertag = "url"
            namespaces = (
                ("a", "http://www.google.com/schemas/sitemap/0.84"),
                ("b", "http://www.example.com/schemas/extras/1.0"),
            )

            def parse_node(self, response, selector):
                yield {
                    "loc": selector.xpath("a:loc/text()").getall(),
                    "updated": selector.xpath("b:updated/text()").getall(),
                    "other": selector.xpath("other/@value").getall(),
                    "custom": selector.xpath("other/@b:custom").getall(),
                }

        for iterator in ("iternodes", "xml"):
            spider = _XMLSpider("example", iterator=iterator)
            output = list(spider._parse(response))
            assert len(output) == 2, iterator
            assert output == [
                {
                    "loc": ["http://www.example.com/Special-Offers.html"],
                    "updated": ["2009-08-16"],
                    "custom": ["fuu"],
                    "other": ["bar"],
                },
                {
                    "loc": [],
                    "updated": ["2009-08-16"],
                    "other": ["foo"],
                    "custom": [],
                },
            ], iterator

    @coroutine_test
    async def test_parse_node_uses_parse_item(self, mockserver: MockServer):
        # parse_node falls back to parse_item for backward compatibility.
        class _Spider(RawFeedSpider, self.spider_class):  # type: ignore[name-defined,misc]
            itertag = "item"

            def raw_body(self):
                return "<items><item><id>1</id></item></items>"

            def parse_item(self, response, selector):
                return {"id": selector.xpath("id/text()").get()}

        items, _ = await crawl_items(_Spider, mockserver)
        assert items == [{"id": "1"}]

    @coroutine_test
    async def test_parse_node_not_defined(self, mockserver: MockServer):
        class _Spider(RawFeedSpider, self.spider_class):  # type: ignore[name-defined,misc]
            itertag = "item"

            def raw_body(self):
                return "<items><item><id>1</id></item></items>"

        items, crawler = await crawl_items(_Spider, mockserver)
        assert items == []
        assert crawler.stats.get_value("spider_exceptions/NotImplementedError") == 1

    @coroutine_test
    async def test_html_iterator(self, mockserver: MockServer):
        class _Spider(RawFeedSpider, self.spider_class):  # type: ignore[name-defined,misc]
            iterator = "html"
            itertag = "item"
            content_type = "text/html"

            def raw_body(self):
                return (
                    "<html><body><item><id>1</id></item>"
                    "<item><id>2</id></item></body></html>"
                )

            def parse_node(self, response, selector):
                return {"id": selector.xpath("id/text()").get()}

        items, _ = await crawl_items(_Spider, mockserver)
        assert items == [{"id": "1"}, {"id": "2"}]

    @coroutine_test
    async def test_unsupported_iterator(self, mockserver: MockServer):
        class _Spider(RawFeedSpider, self.spider_class):  # type: ignore[name-defined,misc]
            iterator = "unsupported"

            def raw_body(self):
                return "<items><item/></items>"

            def parse_node(self, response, selector):
                return {}

        items, crawler = await crawl_items(_Spider, mockserver)
        assert items == []
        assert crawler.stats.get_value("spider_exceptions/NotSupported") == 1

    @pytest.mark.parametrize("feed_iterator", ["xml", "html"])
    @coroutine_test
    async def test_non_text_response(self, feed_iterator: str, mockserver: MockServer):
        # The xml and html iterators require a text response.
        class _Spider(RawFeedSpider, self.spider_class):  # type: ignore[name-defined,misc]
            content_type = "application/octet-stream"
            iterator = feed_iterator

            def raw_body(self):
                # A binary (non-text) body, so the response is a plain Response.
                return "\x00\x01\x02\x03"

            def parse_node(self, response, selector):
                return {}

        items, crawler = await crawl_items(_Spider, mockserver)
        assert items == []
        assert crawler.stats.get_value("spider_exceptions/ValueError") == 1


class TestCSVFeedSpider(TestSpiderBase):
    spider_class = CSVFeedSpider

    def test_parse_rows(self):
        body = get_testdata("feeds", "feed-sample6.csv")
        response = Response("http://example.org/dummy.csv", body=body)

        class _CrawlSpider(self.spider_class):  # type: ignore[name-defined,misc]
            name = "test"
            delimiter = ","
            quotechar = "'"

            def parse_row(self, response, row):
                return row

        spider = _CrawlSpider()
        rows = list(spider.parse_rows(response))
        assert rows[0] == {"id": "1", "name": "alpha", "value": "foobar"}
        assert len(rows) == 4

    @coroutine_test
    async def test_parse(self, mockserver: MockServer):
        class _Spider(RawFeedSpider, self.spider_class):  # type: ignore[name-defined,misc]
            content_type = "text/csv"
            delimiter = ","
            quotechar = "'"

            def raw_body(self):
                return get_testdata("feeds", "feed-sample6.csv").decode()

            def parse_row(self, response, row):
                return row

        items, _ = await crawl_items(_Spider, mockserver)
        assert items[0] == {"id": "1", "name": "alpha", "value": "foobar"}
        assert len(items) == 4

    @coroutine_test
    async def test_parse_row_not_defined(self, mockserver: MockServer):
        class _Spider(RawFeedSpider, self.spider_class):  # type: ignore[name-defined,misc]
            content_type = "text/csv"

            def raw_body(self):
                return "id\n1\n"

        items, crawler = await crawl_items(_Spider, mockserver)
        assert items == []
        assert crawler.stats.get_value("spider_exceptions/NotImplementedError") == 1


class TestNoParseMethodSpider:
    spider_class = Spider

    def test_undefined_parse_method(self):
        spider = self.spider_class("example.com")
        text = b"Random text"
        resp = TextResponse(url="http://www.example.com/random_url", body=text)

        exc_msg = "Spider.parse callback is not defined"
        with pytest.raises(NotImplementedError, match=exc_msg):
            spider.parse(resp)



"""
HTTP basic auth downloader middleware

See documentation in docs/topics/downloader-middleware.rst
"""

from __future__ import annotations

import warnings
from typing import TYPE_CHECKING

from w3lib.http import basic_auth_header

from scrapy import Request, Spider, signals
from scrapy.exceptions import ScrapyDeprecationWarning
from scrapy.settings import SETTINGS_PRIORITIES
from scrapy.utils.decorators import _warn_spider_arg
from scrapy.utils.url import url_is_from_any_domain

if TYPE_CHECKING:
    # typing.Self requires Python 3.11
    from typing_extensions import Self

    from scrapy.crawler import Crawler
    from scrapy.http import Response


class HttpAuthMiddleware:
    """Set Basic HTTP Authorization header."""

    def __init__(self) -> None:
        self._auth: bytes | None = None
        self._domain: str | None = None

    @classmethod
    def from_crawler(cls, crawler: Crawler) -> Self:
        o = cls()
        usr = crawler.settings.get("HTTPAUTH_USER", "")
        pwd = crawler.settings.get("HTTPAUTH_PASS", "")
        if usr or pwd:
            domain_priority = crawler.settings.getpriority("HTTPAUTH_DOMAIN") or 0
            if domain_priority <= SETTINGS_PRIORITIES["default"]:
                raise ValueError(
                    "HTTPAUTH_DOMAIN must be set when HTTPAUTH_USER or HTTPAUTH_PASS "
                    "is configured. Set it to a domain (e.g. 'example.com') to restrict "
                    "credentials to that domain, or set it to None to send credentials "
                    "with all requests."
                )
            o._auth = basic_auth_header(usr, pwd)
            o._domain = crawler.settings.get("HTTPAUTH_DOMAIN")
        crawler.signals.connect(o.spider_opened, signal=signals.spider_opened)
        return o

    def spider_opened(self, spider: Spider) -> None:
        usr = getattr(spider, "http_user", "")
        pwd = getattr(spider, "http_pass", "")
        if usr or pwd:
            warnings.warn(
                "Use the HTTPAUTH_USER, HTTPAUTH_PASS, and HTTPAUTH_DOMAIN settings "
                "instead of the http_user, http_pass, and http_auth_domain spider "
                "attributes. Support for the spider attributes will be removed in a "
                "future version of Scrapy.",
                category=ScrapyDeprecationWarning,
                stacklevel=2,
            )
            self._auth = basic_auth_header(usr, pwd)
            self._domain = spider.http_auth_domain  # type: ignore[attr-defined]

    @_warn_spider_arg
    def process_request(
        self, request: Request, spider: Spider | None = None
    ) -> Request | Response | None:
        if b"Authorization" in request.headers:
            return None
        # Per-request meta overrides
        usr = request.meta.get("http_user", "")
        pwd = request.meta.get("http_pass", "")
        if usr or pwd:
            domain = request.meta.get("http_auth_domain")
            if not domain or url_is_from_any_domain(request.url, [domain]):
                request.headers[b"Authorization"] = basic_auth_header(usr, pwd)
            return None
        # Middleware-level auth
        if self._auth and (
            not self._domain or url_is_from_any_domain(request.url, [self._domain])
        ):
            request.headers[b"Authorization"] = self._auth
        return None



from __future__ import annotations

from html import escape
from pathlib import Path
from typing import TYPE_CHECKING, Any

import pytest

from scrapy import Request
from scrapy.http import HtmlResponse
from scrapy.linkextractors import LinkExtractor
from scrapy.utils.request import fingerprint

if TYPE_CHECKING:
    from pytest_codspeed import BenchmarkFixture  # type: ignore[import-not-found]

pytest.importorskip("pytest_codspeed", reason="Benchmarks require pytest-codspeed")

RESPONSE_URL = "https://www.example.com/catalogue/page-1.html"

# Links that each scenario returns for the benchmark page. They are fewer than
# the anchors of the page because links to images, to other non-crawlable files
# and to non-HTTP schemes are rejected, and, except in the scenario that keeps
# duplicates, because the links that the navigation repeats are collapsed.
LINKS = 63
DUPLICATE_LINKS = 88
CANONICAL_LINKS = 60
FILTERED_LINKS = 45

# Requests built from LINKS links that point to a different resource.
# Canonicalization maps the rest to one that another link already covers, e.g.
# two fragments of a page, or two spellings of one percent-escape.
FINGERPRINTS = 60


def _read_corpus() -> tuple[list[str], list[str]]:
    """Return the URLs of ``urls.txt``, and its first group of URLs.

    The first group is the site navigation, which the benchmark page repeats.
    """
    groups: list[list[str]] = [[]]
    for line in (Path(__file__).parent / "urls.txt").read_text().splitlines():
        line = line.strip()
        if not line or line.startswith("#"):
            if groups[-1]:
                groups.append([])
            continue
        groups[-1].append(line)
    urls = [url for group in groups for url in group]
    return urls, groups[0]


def _build_page(urls: list[str], navigation: list[str]) -> bytes:
    """Return an HTML page that links to *urls*.

    Every link is surrounded by the markup of a product listing, so that
    benchmarks also cover walking over the elements and attributes that a real
    page puts between links.
    """

    def item(index: int, url: str) -> str:
        href = escape(url)
        return (
            f'<li class="product" data-index="{index}">'
            f'<img src="/media/thumbnail-{index}.jpg" alt="Product {index}" '
            f'width="128" height="128">'
            f'<h3><a href="{href}">Product {index}</a></h3>'
            f'<p class="description">A description of product {index}.</p>'
            f"</li>"
        )

    def nav(urls: list[str]) -> str:
        links = "".join(f'<a href="{escape(url)}">{escape(url)}</a>' for url in urls)
        return f'<nav class="site">{links}</nav>'

    items = "".join(item(index, url) for index, url in enumerate(urls))
    return (
        "<!DOCTYPE html><html><head><title>Catalogue</title>"
        f'<base href="{RESPONSE_URL}"></head><body>'
        f'{nav(navigation)}<ul class="products">{items}</ul>{nav(navigation)}'
        "</body></html>"
    ).encode()


URLS, NAVIGATION = _read_corpus()
BODY = _build_page(URLS, NAVIGATION)


def _response() -> HtmlResponse:
    return HtmlResponse(RESPONSE_URL, body=BODY, encoding="utf-8")


@pytest.mark.parametrize(
    ("kwargs", "links"),
    [
        pytest.param({}, LINKS, id="default"),
        pytest.param({"unique": False}, DUPLICATE_LINKS, id="duplicates"),
        pytest.param({"canonicalize": True}, CANONICAL_LINKS, id="canonicalize"),
        pytest.param(
            {
                "allow": r"/catalogue/",
                "deny": r"/legal/",
                "allow_domains": ["example.com", "www.example.com"],
            },
            FILTERED_LINKS,
            id="filtered",
        ),
    ],
)
def test_extract_links(
    benchmark: BenchmarkFixture, kwargs: dict[str, Any], links: int
) -> None:
    """Extraction of every link of a page.

    The scenarios cover the choices that change which work dominates:
    deduplication and canonicalization both build a key for every link, and the
    filters of a configured extractor reject links before the later checks,
    which the default extractor reaches for every link.
    """
    link_extractor = LinkExtractor(**kwargs)

    def run() -> None:
        assert len(link_extractor.extract_links(_response())) == links

    benchmark(run)


EXTRACTED_URLS = [link.url for link in LinkExtractor().extract_links(_response())]


def test_requests(benchmark: BenchmarkFixture) -> None:
    """Building a request for every link of a page."""

    def run() -> None:
        assert len([Request(url) for url in EXTRACTED_URLS]) == LINKS

    benchmark(run)


def test_fingerprints(benchmark: BenchmarkFixture) -> None:
    """Fingerprinting the request of every link of a page.

    Requests are built here as well, and not once for all rounds, because
    fingerprints are cached per request object.
    """

    def run() -> None:
        assert (
            len({fingerprint(Request(url)) for url in EXTRACTED_URLS}) == FINGERPRINTS
        )

    benchmark(run)



from __future__ import annotations

import logging
from typing import TYPE_CHECKING

import pytest

from scrapy import Spider
from scrapy.http import Request
from scrapy.item import Item
from scrapy.settings import Settings
from scrapy.spiderloader import get_spider_loader
from scrapy.utils.spider import (
    iter_spider_classes,
    iterate_spider_output,
    spidercls_for_request,
)

if TYPE_CHECKING:
    from scrapy.spiderloader import SpiderLoaderProtocol


class MySpider1(Spider):
    name = "myspider1"
    allowed_domains = ["example.com", "myspider1.example"]


class MySpider2(Spider):
    name = "myspider2"
    allowed_domains = ["example.com"]


@pytest.fixture
def spider_loader() -> SpiderLoaderProtocol:
    return get_spider_loader(Settings({"SPIDER_MODULES": ["tests.test_utils_spider"]}))


def test_iterate_spider_output():
    i = Item()
    r = Request("http://scrapytest.org")
    o = object()

    assert list(iterate_spider_output(i)) == [i]  # type: ignore[call-overload]
    assert list(iterate_spider_output(r)) == [r]
    assert list(iterate_spider_output(o)) == [o]
    assert list(iterate_spider_output([r, i, o])) == [r, i, o]


def test_iter_spider_classes():
    import tests.test_utils_spider  # noqa: PLW0406,PLC0415

    it = iter_spider_classes(tests.test_utils_spider)
    assert set(it) == {MySpider1, MySpider2}


class TestSpiderclsForRequest:
    def test_single_match(self, spider_loader: SpiderLoaderProtocol) -> None:
        request = Request("http://myspider1.example/")
        assert spidercls_for_request(spider_loader, request) is MySpider1

    def test_no_match(self, spider_loader: SpiderLoaderProtocol) -> None:
        request = Request("http://toscrape.com/")
        assert spidercls_for_request(spider_loader, request) is None
        assert spidercls_for_request(spider_loader, request, MySpider1) is MySpider1

    def test_multiple_matches(self, spider_loader: SpiderLoaderProtocol) -> None:
        request = Request("http://example.com/")
        assert spidercls_for_request(spider_loader, request) is None
        assert spidercls_for_request(spider_loader, request, MySpider2) is MySpider2

    def test_log_none(
        self, spider_loader: SpiderLoaderProtocol, caplog: pytest.LogCaptureFixture
    ) -> None:
        request = Request("http://toscrape.com/")
        with caplog.at_level(logging.ERROR):
            assert spidercls_for_request(spider_loader, request, log_none=True) is None
        assert "Unable to find spider that handles" in caplog.text

    def test_log_multiple(
        self, spider_loader: SpiderLoaderProtocol, caplog: pytest.LogCaptureFixture
    ) -> None:
        request = Request("http://example.com/")
        with caplog.at_level(logging.ERROR):
            assert (
                spidercls_for_request(spider_loader, request, log_multiple=True) is None
            )
        assert "More than one spider can handle" in caplog.text
        assert "myspider1, myspider2" in caplog.text



from scrapy.utils.test import get_crawler
from tests.mockserver.http import MockServer
from tests.spiders import (
    ErrorSpider,
    FollowAllSpider,
    ItemSpider,
    MaxItemsAndRequestsSpider,
    SlowSpider,
)
from tests.utils.decorators import inline_callbacks_test


@inline_callbacks_test
def test_closespider_itemcount(mockserver: MockServer):
    close_on = 5
    crawler = get_crawler(ItemSpider, {"CLOSESPIDER_ITEMCOUNT": close_on})
    yield crawler.crawl(mockserver=mockserver)
    assert isinstance(crawler.spider, ItemSpider)
    reason = crawler.spider.meta["close_reason"]
    assert reason == "closespider_itemcount"
    itemcount = crawler.stats.get_value("item_scraped_count")
    assert itemcount >= close_on


@inline_callbacks_test
def test_closespider_pagecount(mockserver: MockServer):
    close_on = 5
    crawler = get_crawler(FollowAllSpider, {"CLOSESPIDER_PAGECOUNT": close_on})
    yield crawler.crawl(mockserver=mockserver)
    assert isinstance(crawler.spider, FollowAllSpider)
    reason = crawler.spider.meta["close_reason"]
    assert reason == "closespider_pagecount"
    pagecount = crawler.stats.get_value("response_received_count")
    assert pagecount >= close_on


@inline_callbacks_test
def test_closespider_pagecount_no_item(mockserver: MockServer):
    close_on = 5
    max_items = 5
    max_requests = close_on + max_items
    crawler = get_crawler(
        MaxItemsAndRequestsSpider,
        {
            "CLOSESPIDER_PAGECOUNT_NO_ITEM": close_on,
        },
    )
    yield crawler.crawl(
        max_items=max_items, max_requests=max_requests, mockserver=mockserver
    )
    assert isinstance(crawler.spider, MaxItemsAndRequestsSpider)
    reason = crawler.spider.meta["close_reason"]
    assert reason == "closespider_pagecount_no_item"
    pagecount = crawler.stats.get_value("response_received_count")
    itemcount = crawler.stats.get_value("item_scraped_count")
    assert pagecount <= close_on + itemcount


@inline_callbacks_test
def test_closespider_pagecount_no_item_with_pagecount(mockserver: MockServer):
    close_on_pagecount_no_item = 5
    close_on_pagecount = 20
    crawler = get_crawler(
        FollowAllSpider,
        {
            "CLOSESPIDER_PAGECOUNT_NO_ITEM": close_on_pagecount_no_item,
            "CLOSESPIDER_PAGECOUNT": close_on_pagecount,
        },
    )
    yield crawler.crawl(mockserver=mockserver)
    assert isinstance(crawler.spider, FollowAllSpider)
    reason = crawler.spider.meta["close_reason"]
    assert reason == "closespider_pagecount_no_item"
    pagecount = crawler.stats.get_value("response_received_count")
    assert pagecount < close_on_pagecount


@inline_callbacks_test
def test_closespider_errorcount(mockserver: MockServer):
    close_on = 5
    crawler = get_crawler(ErrorSpider, {"CLOSESPIDER_ERRORCOUNT": close_on})
    yield crawler.crawl(total=1000000, mockserver=mockserver)
    assert isinstance(crawler.spider, ErrorSpider)
    reason = crawler.spider.meta["close_reason"]
    assert reason == "closespider_errorcount"
    key = f"spider_exceptions/{crawler.spider.exception_cls.__name__}"
    errorcount = crawler.stats.get_value(key)
    assert crawler.stats.get_value("spider_exceptions/count") >= close_on
    assert errorcount >= close_on


@inline_callbacks_test
def test_closespider_timeout(mockserver: MockServer):
    close_on = 0.1
    crawler = get_crawler(FollowAllSpider, {"CLOSESPIDER_TIMEOUT": close_on})
    yield crawler.crawl(total=1000000, mockserver=mockserver)
    assert isinstance(crawler.spider, FollowAllSpider)
    reason = crawler.spider.meta["close_reason"]
    assert reason == "closespider_timeout"
    total_seconds = crawler.stats.get_value("elapsed_time_seconds")
    assert total_seconds >= close_on


@inline_callbacks_test
def test_closespider_timeout_no_item(mockserver: MockServer):
    timeout = 1
    crawler = get_crawler(SlowSpider, {"CLOSESPIDER_TIMEOUT_NO_ITEM": timeout})
    yield crawler.crawl(n=3, mockserver=mockserver)
    assert isinstance(crawler.spider, SlowSpider)
    reason = crawler.spider.meta["close_reason"]
    assert reason == "closespider_timeout_no_item"
    total_seconds = crawler.stats.get_value("elapsed_time_seconds")
    assert total_seconds >= timeout



"""Tests for scrapy.core.downloader.handlers.http11.HTTP11DownloadHandler."""

from __future__ import annotations

import sys
from typing import TYPE_CHECKING, Any

import pytest

from scrapy import Spider
from scrapy.core.downloader.handlers.http11 import HTTP11DownloadHandler
from scrapy.crawler import Crawler
from scrapy.exceptions import NotConfigured
from scrapy.utils.misc import build_from_crawler
from tests.utils.bases.download_handlers_http import (
    TestHttpBase,
    TestHttpProxyBase,
    TestHttpsBase,
    TestHttpsCustomCiphersBase,
    TestHttpsInvalidDNSIdBase,
    TestHttpsInvalidDNSPatternBase,
    TestHttpsTLSVersionBase,
    TestHttpsWrongHostnameBase,
    TestHttpWithCrawlerBase,
    TestMitmProxyBase,
    TestRealWebsiteBase,
    TestSimpleHttpsBase,
)

if TYPE_CHECKING:
    from scrapy.core.downloader.handlers import DownloadHandlerProtocol


pytestmark = pytest.mark.requires_reactor  # HTTP11DownloadHandler requires a reactor


class HTTP11DownloadHandlerMixin:
    @property
    def download_handler_cls(self) -> type[DownloadHandlerProtocol]:
        return HTTP11DownloadHandler

    @property
    def settings_dict(self) -> dict[str, Any] | None:
        return {
            "DOWNLOAD_HANDLERS": {
                "http": "scrapy.core.downloader.handlers.http11.HTTP11DownloadHandler",
                "https": "scrapy.core.downloader.handlers.http11.HTTP11DownloadHandler",
            }
        }


def test_not_configured_without_reactor() -> None:
    crawler = Crawler(Spider, {"TWISTED_REACTOR_ENABLED": False})
    with pytest.raises(NotConfigured):
        build_from_crawler(HTTP11DownloadHandler, crawler)


class TestHttp(HTTP11DownloadHandlerMixin, TestHttpBase):
    pass


class TestHttps(HTTP11DownloadHandlerMixin, TestHttpsBase):
    pass


class TestSimpleHttps(HTTP11DownloadHandlerMixin, TestSimpleHttpsBase):
    pass


class TestHttpsWrongHostname(HTTP11DownloadHandlerMixin, TestHttpsWrongHostnameBase):
    pass


class TestHttpsInvalidDNSId(HTTP11DownloadHandlerMixin, TestHttpsInvalidDNSIdBase):
    pass


class TestHttpsInvalidDNSPattern(
    HTTP11DownloadHandlerMixin, TestHttpsInvalidDNSPatternBase
):
    pass


class TestHttpsCustomCiphers(HTTP11DownloadHandlerMixin, TestHttpsCustomCiphersBase):
    pass


class TestHttpsTLSVersion(HTTP11DownloadHandlerMixin, TestHttpsTLSVersionBase):
    pass


class TestHttpWithCrawler(HTTP11DownloadHandlerMixin, TestHttpWithCrawlerBase):
    pass


class TestHttpsWithCrawler(TestHttpWithCrawler):
    is_secure = True


class TestHttpProxy(HTTP11DownloadHandlerMixin, TestHttpProxyBase):
    pass


class TestHttpsProxy(HTTP11DownloadHandlerMixin, TestHttpProxyBase):
    is_secure = True
    # not implemented
    handler_supports_tls_in_tls = False


@pytest.mark.requires_mitmproxy
class TestMitmProxy(HTTP11DownloadHandlerMixin, TestMitmProxyBase):
    # not implemented
    handler_supports_tls_in_tls = False


@pytest.mark.requires_internet
class TestRealWebsite(HTTP11DownloadHandlerMixin, TestRealWebsiteBase):
    @property
    def platform_cert_store_works(self) -> bool:
        return sys.platform != "win32"



from __future__ import annotations

from abc import ABC
from typing import TYPE_CHECKING

from .base import BaseDownloadHandler

if TYPE_CHECKING:
    from scrapy.crawler import Crawler


class BaseHttpDownloadHandler(BaseDownloadHandler, ABC):
    """Base class for built-in HTTP download handlers."""

    def __init__(self, crawler: Crawler):
        super().__init__(crawler)
        self._default_maxsize: int = crawler.settings.getint("DOWNLOAD_MAXSIZE")
        self._default_warnsize: int = crawler.settings.getint("DOWNLOAD_WARNSIZE")
        self._fail_on_dataloss: bool = crawler.settings.getbool(
            "DOWNLOAD_FAIL_ON_DATALOSS"
        )
        self._tls_verbose_logging: bool = crawler.settings.getbool(
            "DOWNLOADER_CLIENT_TLS_VERBOSE_LOGGING"
        )
        self._fail_on_dataloss_warned: bool = False



from __future__ import annotations

import hashlib
import logging
import shutil
import sys
import tempfile
from pathlib import Path
from typing import TYPE_CHECKING, Any

import pytest

from scrapy.core.scheduler import Scheduler
from scrapy.dupefilters import BaseDupeFilter, RFPDupeFilter
from scrapy.exceptions import ScrapyDeprecationWarning
from scrapy.http import Request
from scrapy.utils.misc import build_from_crawler
from scrapy.utils.python import to_bytes
from scrapy.utils.test import get_crawler
from tests.spiders import SimpleSpider

if TYPE_CHECKING:
    from scrapy.crawler import Crawler


def _get_dupefilter(
    *,
    crawler: Crawler | None = None,
    settings: dict[str, Any] | None = None,
    open_: bool = True,
) -> BaseDupeFilter:
    if crawler is None:
        crawler = get_crawler(settings_dict=settings)
    scheduler = build_from_crawler(Scheduler, crawler)
    dupefilter = scheduler.df
    if open_:
        dupefilter.open()
    return dupefilter


class FromCrawlerRFPDupeFilter(RFPDupeFilter):
    @classmethod
    def from_crawler(cls, crawler):
        df = super().from_crawler(crawler)
        df.method = "from_crawler"
        return df


class DirectDupeFilter:
    method = "n/a"


class TestRFPDupeFilter:
    def test_df_from_crawler_scheduler(self):
        settings = {
            "DUPEFILTER_DEBUG": True,
            "DUPEFILTER_CLASS": FromCrawlerRFPDupeFilter,
        }
        crawler = get_crawler(settings_dict=settings)
        scheduler = build_from_crawler(Scheduler, crawler)
        assert scheduler.df.debug
        assert scheduler.df.method == "from_crawler"

    def test_df_direct_scheduler(self):
        settings = {
            "DUPEFILTER_CLASS": DirectDupeFilter,
        }
        crawler = get_crawler(settings_dict=settings)
        scheduler = build_from_crawler(Scheduler, crawler)
        assert scheduler.df.method == "n/a"

    def test_filter(self):
        dupefilter = _get_dupefilter()
        r1 = Request("http://scrapytest.org/1")
        r2 = Request("http://scrapytest.org/2")
        r3 = Request("http://scrapytest.org/2")

        assert not dupefilter.request_seen(r1)
        assert dupefilter.request_seen(r1)

        assert not dupefilter.request_seen(r2)
        assert dupefilter.request_seen(r3)

        dupefilter.close("finished")

    def test_dupefilter_path(self):
        r1 = Request("http://scrapytest.org/1")
        r2 = Request("http://scrapytest.org/2")

        path = tempfile.mkdtemp()
        try:
            df = _get_dupefilter(settings={"JOBDIR": path}, open_=False)
            try:
                df.open()
                assert not df.request_seen(r1)
                assert df.request_seen(r1)
            finally:
                df.close("finished")

            df2 = _get_dupefilter(settings={"JOBDIR": path}, open_=False)
            assert df is not df2
            try:
                df2.open()
                assert df2.request_seen(r1)
                assert not df2.request_seen(r2)
                assert df2.request_seen(r2)
            finally:
                df2.close("finished")
        finally:
            shutil.rmtree(path)

    def test_request_fingerprint(self):
        """Test if customization of request_fingerprint method will change
        output of request_seen.

        """
        dupefilter = _get_dupefilter()
        r1 = Request("http://scrapytest.org/index.html")
        r2 = Request("http://scrapytest.org/INDEX.html")

        assert not dupefilter.request_seen(r1)
        assert not dupefilter.request_seen(r2)

        dupefilter.close("finished")

        class RequestFingerprinter:
            def fingerprint(self, request):
                fp = hashlib.sha1()
                fp.update(to_bytes(request.url.lower()))
                return fp.digest()

        settings = {"REQUEST_FINGERPRINTER_CLASS": RequestFingerprinter}
        case_insensitive_dupefilter = _get_dupefilter(settings=settings)

        assert not case_insensitive_dupefilter.request_seen(r1)
        assert case_insensitive_dupefilter.request_seen(r2)

        case_insensitive_dupefilter.close("finished")

    def test_seenreq_newlines(self):
        r"""Checks against adding duplicate \r to
        line endings on Windows platforms."""

        r1 = Request("http://scrapytest.org/1")

        path = tempfile.mkdtemp()
        crawler = get_crawler(settings_dict={"JOBDIR": path})
        try:
            scheduler = build_from_crawler(Scheduler, crawler)
            df = scheduler.df
            df.open()
            df.request_seen(r1)
            df.close("finished")

            with Path(path, "requests.seen").open("rb") as seen_file:
                line = next(seen_file).decode()
                assert not line.endswith("\r\r\n")
                if sys.platform == "win32":
                    assert line.endswith("\r\n")
                else:
                    assert line.endswith("\n")

        finally:
            shutil.rmtree(path)

    def test_log(self, caplog: pytest.LogCaptureFixture) -> None:
        settings = {
            "DUPEFILTER_DEBUG": False,
            "DUPEFILTER_CLASS": FromCrawlerRFPDupeFilter,
        }
        crawler = get_crawler(SimpleSpider, settings_dict=settings)
        spider = SimpleSpider.from_crawler(crawler)
        dupefilter = _get_dupefilter(crawler=crawler)

        r1 = Request("http://scrapytest.org/index.html")
        r2 = Request("http://scrapytest.org/index.html")

        with caplog.at_level(logging.DEBUG):
            dupefilter.log(r1, spider)
            dupefilter.log(r2, spider)

        assert crawler.stats.get_value("dupefilter/filtered") == 2
        assert (
            "scrapy.dupefilters",
            logging.DEBUG,
            "Filtered duplicate request: <GET http://scrapytest.org/index.html> - no more"
            " duplicates will be shown (see DUPEFILTER_DEBUG to show all duplicates)",
        ) in caplog.record_tuples

        dupefilter.close("finished")

    @pytest.mark.parametrize("df", [None, FromCrawlerRFPDupeFilter])
    def test_log_debug(
        self, caplog: pytest.LogCaptureFixture, df: type[BaseDupeFilter] | None
    ) -> None:
        settings: dict[str, Any] = {
            "DUPEFILTER_DEBUG": True,
        }
        if df:
            settings["DUPEFILTER_CLASS"] = df
        crawler = get_crawler(SimpleSpider, settings_dict=settings)
        spider = SimpleSpider.from_crawler(crawler)
        dupefilter = _get_dupefilter(crawler=crawler)

        r1 = Request("http://scrapytest.org/index.html")
        r2 = Request(
            "http://scrapytest.org/index.html",
            headers={"Referer": "http://scrapytest.org/INDEX.html"},
        )

        with caplog.at_level(logging.DEBUG):
            dupefilter.log(r1, spider)
            dupefilter.log(r2, spider)

        assert crawler.stats.get_value("dupefilter/filtered") == 2
        assert (
            "scrapy.dupefilters",
            logging.DEBUG,
            "Filtered duplicate request: <GET http://scrapytest.org/index.html> (referer: None)",
        ) in caplog.record_tuples
        assert (
            "scrapy.dupefilters",
            logging.DEBUG,
            "Filtered duplicate request: <GET http://scrapytest.org/index.html>"
            " (referer: http://scrapytest.org/INDEX.html)",
        ) in caplog.record_tuples

        dupefilter.close("finished")


class TestBaseDupeFilter:
    def test_log_deprecation(self):
        dupefilter = _get_dupefilter(
            settings={"DUPEFILTER_CLASS": BaseDupeFilter},
        )
        with pytest.warns(
            ScrapyDeprecationWarning,
            match=r"Calling BaseDupeFilter\.log\(\) is deprecated.",
        ):
            dupefilter.log(None, None)



from __future__ import annotations

from scrapy.downloadermiddlewares.useragent import UserAgentMiddleware
from scrapy.http import Request
from scrapy.spiders import Spider
from scrapy.utils.misc import build_from_crawler
from scrapy.utils.test import get_crawler


def get_spider_and_mw(
    default_useragent: str | None,
) -> tuple[Spider, UserAgentMiddleware]:
    crawler = get_crawler(Spider, {"USER_AGENT": default_useragent})
    spider = crawler._create_spider("foo")
    return spider, build_from_crawler(UserAgentMiddleware, crawler)


def test_default_agent():
    _, mw = get_spider_and_mw("default_useragent")
    req = Request("http://scrapytest.org/")
    assert mw.process_request(req) is None
    assert req.headers["User-Agent"] == b"default_useragent"


def test_header_agent():
    spider, mw = get_spider_and_mw("default_useragent")
    mw.spider_opened(spider)
    req = Request("http://scrapytest.org/", headers={"User-Agent": "header_useragent"})
    assert mw.process_request(req) is None
    assert req.headers["User-Agent"] == b"header_useragent"


def test_no_agent():
    spider, mw = get_spider_and_mw(None)
    mw.spider_opened(spider)
    req = Request("http://scrapytest.org/")
    assert mw.process_request(req) is None
    assert "User-Agent" not in req.headers



from __future__ import annotations

from typing import Any

from scrapy.http.request import NO_CALLBACK, Request
from scrapy.pipelines.media import MediaPipeline
from scrapy.utils.spider import DefaultSpider

# required by persist_file() and stat_file(), but as some stores don't use the argument
# we can pass this singleton to keep type hints correct
DUMMY_SPIDER_INFO = MediaPipeline.SpiderInfo(DefaultSpider())


async def mocked_download_func(request: Request) -> Any:
    assert request.callback is NO_CALLBACK
    response = request.meta.get("response")
    if callable(response):
        response = await response()
    if isinstance(response, Exception):
        raise response
    return response



from __future__ import annotations

import logging
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any
from unittest import mock

import pytest

from scrapy import signals
from scrapy.crawler import Crawler
from scrapy.exceptions import ScrapyDeprecationWarning
from scrapy.settings import Settings
from scrapy.utils.test import get_crawler, get_reactor_settings
from tests.utils.decorators import inline_callbacks_test

if TYPE_CHECKING:
    from scrapy.spiders import Spider


class TestSpiderBase(ABC):
    @property
    @abstractmethod
    def spider_class(self) -> type[Spider]:
        raise NotImplementedError

    def test_base_spider(self):
        spider = self.spider_class("example.com")
        assert spider.name == "example.com"
        assert spider.start_urls == []

    def test_spider_args(self):
        """``__init__`` method arguments are assigned to spider attributes"""
        spider = self.spider_class("example.com", foo="bar")
        assert spider.foo == "bar"  # type: ignore[attr-defined]

    def test_spider_without_name(self):
        """``__init__`` raises when the name is not provided."""
        msg = "must have a name"
        with pytest.raises(ValueError, match=msg):
            self.spider_class()
        with pytest.raises(ValueError, match=msg):
            self.spider_class(somearg="foo")

    def test_from_crawler_crawler_and_settings_population(self):
        crawler = get_crawler()
        spider = self.spider_class.from_crawler(crawler, "example.com")
        assert hasattr(spider, "crawler")
        assert spider.crawler is crawler
        assert hasattr(spider, "settings")
        assert spider.settings is crawler.settings

    def test_from_crawler_init_call(self):
        with mock.patch.object(
            self.spider_class, "__init__", return_value=None
        ) as mock_init:
            self.spider_class.from_crawler(get_crawler(), "example.com", foo="bar")
            mock_init.assert_called_once_with("example.com", foo="bar")

    def test_closed_signal_call(self):
        class TestSpider(self.spider_class):  # type: ignore[name-defined,misc]
            closed_called = False

            def closed(self, reason):
                self.closed_called = True

        crawler = get_crawler()
        spider = TestSpider.from_crawler(crawler, "example.com")
        crawler.signals.send_catch_log(signal=signals.spider_opened, spider=spider)
        crawler.signals.send_catch_log(
            signal=signals.spider_closed, spider=spider, reason=None
        )
        assert spider.closed_called

    def test_update_settings(self):
        spider_settings = {"TEST1": "spider", "TEST2": "spider"}
        project_settings = {"TEST1": "project", "TEST3": "project"}
        self.spider_class.custom_settings = spider_settings
        settings = Settings(project_settings, priority="project")

        self.spider_class.update_settings(settings)
        assert settings.get("TEST1") == "spider"
        assert settings.get("TEST2") == "spider"
        assert settings.get("TEST3") == "project"

    @inline_callbacks_test
    def test_settings_in_from_crawler(self):
        spider_settings = {"TEST1": "spider", "TEST2": "spider"}
        project_settings = {
            "TEST1": "project",
            "TEST3": "project",
            **get_reactor_settings(),
        }

        class TestSpider(self.spider_class):  # type: ignore[name-defined,misc]
            name = "test"
            custom_settings = spider_settings

            @classmethod
            def from_crawler(cls, crawler: Crawler, *args: Any, **kwargs: Any):
                spider = super().from_crawler(crawler, *args, **kwargs)
                spider.settings.set("TEST1", "spider_instance", priority="spider")
                return spider

        crawler = Crawler(TestSpider, project_settings)
        assert crawler.settings.get("TEST1") == "spider"
        assert crawler.settings.get("TEST2") == "spider"
        assert crawler.settings.get("TEST3") == "project"
        yield crawler.crawl()
        assert crawler.settings.get("TEST1") == "spider_instance"

    def test_logger(self, caplog: pytest.LogCaptureFixture) -> None:
        spider = self.spider_class("example.com")
        caplog.clear()
        with caplog.at_level(logging.INFO):
            spider.logger.info("test log msg")
        assert caplog.record_tuples == [("example.com", logging.INFO, "test log msg")]

        record = caplog.records[0]
        assert getattr(record, "spider", None) is spider

    def test_log(self):
        spider = self.spider_class("example.com")
        with (
            mock.patch("scrapy.spiders.Spider.logger") as mock_logger,
            pytest.warns(
                ScrapyDeprecationWarning, match=r"Spider.log\(\) is deprecated"
            ),
        ):
            spider.log("test log msg", "INFO")  # type: ignore[arg-type]
        mock_logger.log.assert_called_once_with("INFO", "test log msg")



"""
This module implements the :class:`XmlResponse` class which is used as a
content type marker by :class:`~scrapy.selector.Selector` and can be used in
``isinstance()`` checks.

See documentation in docs/topics/request-response.rst
"""

from scrapy.http.response.text import TextResponse


class XmlResponse(TextResponse):
    __slots__ = ()



from __future__ import annotations

import sys
import warnings
from typing import TYPE_CHECKING, Any

import pytest
from twisted.internet.defer import Deferred

from scrapy.exceptions import ScrapyDeprecationWarning
from scrapy.utils.decorators import _warn_spider_arg, deprecated, inthread
from scrapy.utils.defer import maybe_deferred_to_future
from tests.utils.decorators import coroutine_test

if TYPE_CHECKING:
    from collections.abc import AsyncGenerator, Callable


class TestDeprecated:
    def test_warns_and_still_calls(self):
        @deprecated()
        def add(a: int, b: int) -> int:
            return a + b

        with pytest.warns(
            ScrapyDeprecationWarning, match=r"Call to deprecated function add\."
        ):
            result = add(2, 3)

        assert result == 5

    def test_use_instead_in_message(self):
        @deprecated(use_instead="other_function")
        def old() -> None:
            return None

        with pytest.warns(
            ScrapyDeprecationWarning,
            match=r"Call to deprecated function old\. Use other_function instead\.",
        ):
            old()

    def test_applied_without_parentheses(self):
        @deprecated
        def square(x: int) -> int:
            return x * x

        with pytest.warns(
            ScrapyDeprecationWarning, match=r"Call to deprecated function square\."
        ) as record:
            result = square(4)

        assert result == 16
        # No "Use ... instead." part when applied directly to the function.
        assert "instead" not in str(record[0].message)


class TestInthread:
    @coroutine_test
    async def test_returns_deferred_with_result(self):
        @inthread
        def multiply(a, b):
            return a * b

        deferred = multiply(6, 7)
        assert isinstance(deferred, Deferred)
        assert await maybe_deferred_to_future(deferred) == 42


class TestWarnSpiderArg:
    def test_sync_warns_with_spider_arg(self):
        @_warn_spider_arg
        def parse(response: str, spider: str | None = None) -> str:
            return response

        with pytest.warns(
            ScrapyDeprecationWarning, match=r"Passing a 'spider' argument"
        ):
            assert parse("response", spider="spider") == "response"

    @pytest.mark.skipif(
        sys.version_info < (3, 14),
        reason="annotations are only lazily evaluated since Python 3.14 (PEP 649)",
    )
    def test_sync_warns_with_unresolvable_annotations(self):
        # dont_inherit=True, or the module's future import stringizes the annotations
        namespace: dict[str, Any] = {}
        exec(  # pylint: disable=exec-used
            compile(
                "def parse(response: OnlyAtTypeCheckingTime,"
                " spider: OnlyAtTypeCheckingTime | None = None): return response",
                "<test>",
                "exec",
                dont_inherit=True,
            ),
            namespace,
        )
        parse_func: Callable[..., str] = namespace["parse"]
        parse = _warn_spider_arg(parse_func)

        with pytest.warns(
            ScrapyDeprecationWarning, match=r"Passing a 'spider' argument"
        ):
            assert parse("response", spider="spider") == "response"

    def test_sync_warns_with_positional_spider_arg(self):
        @_warn_spider_arg
        def parse(response: str, spider: str | None = None) -> str:
            return response

        with pytest.warns(
            ScrapyDeprecationWarning, match=r"Passing a 'spider' argument"
        ):
            assert parse("response", "spider") == "response"

    def test_sync_warns_with_keyword_only_spider_arg(self):
        @_warn_spider_arg
        def parse(response: str, *, spider: str | None = None) -> str:
            return response

        with pytest.warns(
            ScrapyDeprecationWarning, match=r"Passing a 'spider' argument"
        ):
            assert parse("response", spider="spider") == "response"

    def test_sync_no_warning_without_spider_arg(self):
        @_warn_spider_arg
        def parse(response: str, spider: str | None = None) -> str:
            return response

        with warnings.catch_warnings():
            warnings.simplefilter("error", category=ScrapyDeprecationWarning)
            assert parse("response") == "response"

    @coroutine_test
    async def test_async_warns_with_spider_arg(self):
        @_warn_spider_arg
        async def parse(response: str, spider: str | None = None) -> str:
            return response

        with pytest.warns(
            ScrapyDeprecationWarning, match=r"Passing a 'spider' argument"
        ):
            assert await parse("response", spider="spider") == "response"

    @coroutine_test
    async def test_asyncgen_warns_with_spider_arg(self):
        @_warn_spider_arg
        async def parse(
            response: str, spider: str | None = None
        ) -> AsyncGenerator[str]:
            yield response

        with pytest.warns(
            ScrapyDeprecationWarning, match=r"Passing a 'spider' argument"
        ):
            results = [item async for item in parse("response", spider="spider")]

        assert results == ["response"]
